from dagster import Enum, Field, IntSource, Permissive, Shape

from dagster_cloud.opentelemetry.enum import MetricsReaderEnum

metric_reader_enum = Enum.from_python_enum_direct_values(MetricsReaderEnum)


def metric_reader_config_schema() -> Field:
    return Field(
        config=Shape(
            description="Provide configuration for an OpenTelemetry metric reader.",
            fields={
                "type": Field(
                    config=metric_reader_enum,
                    description="The type of reader to use to capture and export metrics.",
                    is_required=False,
                ),
                "params": Field(
                    description="Provide custom parameters for the OpenTelemetry metric reader.",
                    is_required=False,
                    config=Permissive(
                        description="Provide configuration for the OpenTelemetry metric reader.",
                        fields={
                            "export_interval_millis": Field(
                                IntSource,
                                description="The interval in milliseconds at which to export metrics.",
                                is_required=False,
                            ),
                            "export_timeout_millis": Field(
                                IntSource,
                                description="The timeout in milliseconds for exporting metrics.",
                                is_required=False,
                            ),
                        },
                    ),
                ),
            },
        )
    )



from typing import Dict, List

from dagster import DagsterError
from dagster._utils.error import SerializableErrorInfo
from grpc import Call as GrpcCall


def extract_dagster_error_attributes(error: DagsterError) -> Dict[str, str]:
    """Extracts attributes from a DagsterError that can be used for logging or metrics."""
    error_attributes = {
        "exception": type(error).__name__,
    }

    # DagsterError subtypes often have a field that contains a SerializableErrorInfo object
    # which identify the cause of the error. This field names vary between error types and is not
    # always present, but if they are they often identify the root cause of the error.
    for field in dir(error):
        if field.startswith("__"):
            # Skip magic methods and fields
            continue

        attr = getattr(error, field)
        if type(attr) is list and all(isinstance(item, SerializableErrorInfo) for item in attr):
            serializable_error_infos: List[SerializableErrorInfo] = getattr(error, field)
            for serializable_error_info in serializable_error_infos:
                if serializable_error_info.cause:
                    error_attributes["cause"] = serializable_error_info.cause.cls_name
                    return error_attributes

    # If not obtained from SerializableErrorInfo, use __cause__ to determine the error attributes
    if hasattr(error, "__cause__") and error.__cause__:
        error_cause = error.__cause__
        cause_name = type(error_cause).__name__
        if cause_name:
            error_attributes["cause"] = cause_name

        if isinstance(error_cause, GrpcCall):
            grpc_call: GrpcCall = error_cause
            code = grpc_call.code()
            error_attributes["code"] = code.name

    return error_attributes



from typing import Any, List, Mapping, Optional

from dagster import AssetKey, JobDefinition

OUTPUT_NON_ASSET_SIGIL = "__bigquery_query_metadata_"
BIGQUERY_METADATA_BYTES_BILLED = "__bigquery_bytes_billed"
BIGQUERY_METADATA_SLOTS_MS = "__bigquery_slots_ms"
BIGQUERY_METADATA_JOB_IDS = "__bigquery_job_ids"


def marker_asset_key_for_job(
    job: JobDefinition,
) -> AssetKey:
    return AssetKey(path=[f"{OUTPUT_NON_ASSET_SIGIL}{job.name}"])


def build_bigquery_cost_metadata(
    job_ids: Optional[List[str]], bytes_billed: int, slots_ms: int
) -> Mapping[str, Any]:
    metadata: Mapping[str, Any] = {
        BIGQUERY_METADATA_BYTES_BILLED: bytes_billed,
        BIGQUERY_METADATA_SLOTS_MS: slots_ms,
    }
    if job_ids:
        metadata[BIGQUERY_METADATA_JOB_IDS] = job_ids
    return metadata



import datetime
import io
import os
import pickle
from typing import Any, Tuple, Union

import boto3
import dagster._check as check
import requests
from dagster import InputContext, OutputContext, UPathIOManager, io_manager
from dagster._utils import PICKLE_PROTOCOL
from dagster._vendored.dateutil import parser
from upath import UPath

ECS_AGENT_IP = "169.254.170.2"


class PickledObjectServerlessIOManager(UPathIOManager):
    def __init__(
        self,
        s3_bucket,
        s3_prefix,
    ):
        self._bucket = check.str_param(s3_bucket, "s3_bucket")
        self._s3_prefix = check.str_param(s3_prefix, "s3_prefix")
        self._boto_session, self._boto_session_expiration = self._refresh_boto_session()
        base_path = UPath(s3_prefix) if s3_prefix else None
        super().__init__(base_path=base_path)

    def _refresh_boto_session(self) -> Tuple[boto3.Session, datetime.datetime]:
        # We have to do this whacky way to get credentials to ensure that we get iam role
        # we assigned to the task. If we used the default boto behavior, it could get overriden
        # when users set AWS env vars.
        relative_uri = os.environ["AWS_CONTAINER_CREDENTIALS_RELATIVE_URI"]
        aws_creds = requests.get(f"http://{ECS_AGENT_IP}{relative_uri}").json()
        session = boto3.Session(
            aws_access_key_id=aws_creds["AccessKeyId"],
            aws_secret_access_key=aws_creds["SecretAccessKey"],
            aws_session_token=aws_creds["Token"],
        )
        expiration = parser.parse(aws_creds["Expiration"])
        return session, expiration  # pyright: ignore[reportReturnType]

    @property
    def _s3(self):
        if self._boto_session_expiration <= datetime.datetime.now(
            self._boto_session_expiration.tzinfo
        ) + datetime.timedelta(minutes=5):
            self._boto_session, self._boto_session_expiration = self._refresh_boto_session()
        return self._boto_session.client("s3", region_name="us-west-2")

    def load_from_path(self, context: InputContext, path: UPath) -> Any:
        try:
            s3_obj = self._s3.get_object(Bucket=self._bucket, Key=path.as_posix())["Body"].read()
            return pickle.loads(s3_obj)
        except self._s3.exceptions.NoSuchKey:
            raise FileNotFoundError(
                f"Could not find the input for '{context.name}'. It may have expired."
            )

    def dump_to_path(self, context: OutputContext, obj: Any, path: UPath) -> None:
        pickled_obj = pickle.dumps(obj, PICKLE_PROTOCOL)
        pickled_obj_bytes = io.BytesIO(pickled_obj)
        self._s3.upload_fileobj(pickled_obj_bytes, self._bucket, path.as_posix())

    def path_exists(self, path: UPath) -> bool:
        try:
            self._s3.get_object(Bucket=self._bucket, Key=path.as_posix())
        except self._s3.exceptions.NoSuchKey:
            return False
        return True

    def unlink(self, path: UPath) -> None:
        self._s3.delete_object(Bucket=self._bucket, Key=path.as_posix())

    def make_directory(self, path: UPath) -> None:
        # It is not necessary to create directories in S3
        return None

    def get_op_output_relative_path(self, context: Union[InputContext, OutputContext]):
        from upath import UPath

        return UPath(*["storage", *context.get_identifier()])


@io_manager
def serverless_io_manager(init_context):
    bucket = os.getenv("DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_BUCKET")
    prefix = os.getenv("DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_PREFIX")
    check.invariant(
        bucket and prefix,
        "The serverless_io_manager is only supported when running on Dagster Cloud Serverless."
        " DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_BUCKET or"
        " DAGSTER_CLOUD_SERVERLESS_STORAGE_S3_PREFIX not found.",
    )

    deployment_name = init_context.instance.deployment_name

    return PickledObjectServerlessIOManager(
        bucket, s3_prefix=f"{prefix}/io_storage/{deployment_name}"
    )



import logging
from typing import Dict

from dagster import DagsterRun


def extract_run_attributes(deployment_name, run: DagsterRun) -> Dict[str, str]:
    attributes = {
        "job": run.job_name or "unknown",
        "repository": "unknown",
        "location": "unknown",
        "deployment": deployment_name,
    }

    try:
        if run.remote_job_origin:
            attributes["repository"] = (
                run.remote_job_origin.repository_origin.repository_name or "unknown"
            )
            attributes["location"] = run.remote_job_origin.location_name or "unknown"
    except Exception as e:  # pylint: disable=broad-except
        logging.warning("Failed to extract additional run attributes", exc_info=e)
    finally:
        return attributes



"""Files in this directory are generated by running:

    python -m dagster_cloud.pex.grpc.compile
"""

from . import multi_pex_api_pb2 as multi_pex_api_pb2
from .multi_pex_api_pb2_grpc import (
    MultiPexApiServicer as MultiPexApiServicer,
    MultiPexApiStub as MultiPexApiStub,
)



import os
import sys
import time
from contextlib import contextmanager

import dagster._check as check
import grpc
from dagster._core.errors import DagsterUserCodeProcessError, DagsterUserCodeUnreachableError
from dagster._grpc.client import DEFAULT_GRPC_TIMEOUT
from dagster._grpc.utils import max_rx_bytes, max_send_bytes
from dagster._serdes import serialize_value
from dagster._serdes.serdes import deserialize_value
from dagster._utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info

from .__generated__ import MultiPexApiStub, multi_pex_api_pb2
from .types import (
    CreatePexServerArgs,
    CreatePexServerResponse,
    GetCrashedPexServersArgs,
    GetCrashedPexServersResponse,
    GetPexServersArgs,
    GetPexServersResponse,
    ShutdownPexServerArgs,
    ShutdownPexServerResponse,
)


class MultiPexGrpcClient:
    def __init__(self, port=None, socket=None, host="localhost"):
        self.port = check.opt_int_param(port, "port")
        self.socket = check.opt_str_param(socket, "socket")
        self.host = check.opt_str_param(host, "host")
        if port:
            self._server_address = host + ":" + str(port)
        else:
            self._server_address = "unix:" + os.path.abspath(socket)  # pyright: ignore[reportArgumentType,reportCallIssue]

    def create_pex_server(self, create_pex_server_args: CreatePexServerArgs):
        check.inst_param(create_pex_server_args, "create_pex_server_args", CreatePexServerArgs)
        res = self._query(
            "CreatePexServer",
            multi_pex_api_pb2.CreatePexServerRequest,
            create_pex_server_args=serialize_value(create_pex_server_args),
        )
        return self._response_or_error(res.create_pex_server_response, CreatePexServerResponse)

    def get_pex_servers(self, get_pex_servers_args: GetPexServersArgs) -> GetPexServersResponse:
        check.inst_param(get_pex_servers_args, "get_pex_servers_args", GetPexServersArgs)
        res = self._query(
            "GetPexServers",
            multi_pex_api_pb2.GetPexServersRequest,
            get_pex_servers_args=serialize_value(get_pex_servers_args),
        )
        return self._response_or_error(res.get_pex_servers_response, GetPexServersResponse)

    def get_crashed_pex_servers(
        self, get_crashed_pex_servers_args: GetCrashedPexServersArgs
    ) -> GetCrashedPexServersResponse:
        check.inst_param(
            get_crashed_pex_servers_args, "get_crashed_pex_servers_args", GetCrashedPexServersArgs
        )
        res = self._query(
            "GetCrashedPexServers",
            multi_pex_api_pb2.GetCrashedPexServersRequest,
            get_crashed_pex_servers_args=serialize_value(get_crashed_pex_servers_args),
        )
        return self._response_or_error(
            res.get_crashed_pex_servers_response, GetCrashedPexServersResponse
        )

    def shutdown_pex_server(self, shutdown_pex_server_args: ShutdownPexServerArgs):
        check.inst_param(
            shutdown_pex_server_args, "shutdown_pex_server_args", ShutdownPexServerArgs
        )
        res = self._query(
            "ShutdownPexServer",
            multi_pex_api_pb2.ShutdownPexServerRequest,
            shutdown_pex_server_args=serialize_value(shutdown_pex_server_args),
        )
        return self._response_or_error(res.shutdown_pex_server_response, ShutdownPexServerResponse)

    @contextmanager
    def _channel(self):
        options = [
            ("grpc.max_receive_message_length", max_rx_bytes()),
            ("grpc.max_send_message_length", max_send_bytes()),
        ]
        with grpc.insecure_channel(
            self._server_address,
            options=options,
            compression=grpc.Compression.Gzip,
        ) as channel:
            yield channel

    def _response_or_error(self, response, response_type):
        deserialized_response = deserialize_value(response, (response_type, SerializableErrorInfo))
        if isinstance(deserialized_response, SerializableErrorInfo):
            raise DagsterUserCodeProcessError.from_error_info(deserialized_response)
        check.invariant(isinstance(deserialized_response, response_type))
        return deserialized_response

    def _query(self, method, request_type, timeout=DEFAULT_GRPC_TIMEOUT, **kwargs):
        try:
            with self._channel() as channel:
                stub = MultiPexApiStub(channel)
                response = getattr(stub, method)(request_type(**kwargs), timeout=timeout)
            return response
        except Exception as e:
            raise DagsterUserCodeUnreachableError("Could not reach user code server") from e

    def ping(self, echo):
        check.str_param(echo, "echo")
        res = self._query("Ping", multi_pex_api_pb2.PingRequest, echo=echo)
        return res.echo


def wait_for_grpc_server(client, timeout=180):
    start_time = time.time()

    while True:
        try:
            client.ping("")
            return
        except Exception:
            last_error = serializable_error_info_from_exc_info(sys.exc_info())

        if time.time() - start_time > timeout:
            raise Exception(
                f"Timed out after waiting {timeout}s for server. "
                f"Most recent connection error: {last_error}"
            )

        time.sleep(1)



import logging
import os
import sys
import time
from collections import defaultdict, deque
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import ExitStack
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    DefaultDict,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

import dagster._check as check
from dagster import DagsterInstance
from dagster._core.launcher.base import LaunchRunContext
from dagster._core.remote_representation import CodeLocationOrigin
from dagster._core.remote_representation.origin import RegisteredCodeLocationOrigin
from dagster._core.utils import FuturesAwareThreadPoolExecutor
from dagster._grpc.client import DagsterGrpcClient
from dagster._grpc.types import CancelExecutionRequest
from dagster._serdes import deserialize_value, serialize_value
from dagster._time import get_current_datetime, get_current_timestamp
from dagster._utils.cached_method import cached_method
from dagster._utils.container import retrieve_containerized_utilization_metrics
from dagster._utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info
from dagster._utils.interrupts import raise_interrupts_as
from dagster._utils.merger import merge_dicts
from dagster._utils.typed_dict import init_optional_typeddict
from dagster_cloud_cli.core.errors import DagsterCloudHTTPError, raise_http_error
from dagster_cloud_cli.core.workspace import CodeLocationDeployData, get_instance_ref_for_user_code

from dagster_cloud.api.dagster_cloud_api import (
    AgentHeartbeat,
    AgentUtilizationMetrics,
    BatchDagsterCloudUploadApiResponse,
    DagsterCloudApi,
    DagsterCloudApiErrorResponse,
    DagsterCloudApiGrpcResponse,
    DagsterCloudApiRequest,
    DagsterCloudApiResponse,
    DagsterCloudApiSuccess,
    DagsterCloudApiThreadTelemetry,
    DagsterCloudApiUnknownCommandResponse,
    DagsterCloudUploadApiResponse,
    TimestampedError,
)
from dagster_cloud.batching import Batcher
from dagster_cloud.instance import DagsterCloudAgentInstance
from dagster_cloud.util.errors import truncate_serialized_error
from dagster_cloud.workspace.user_code_launcher import (
    DagsterCloudUserCodeLauncher,
    UserCodeLauncherEntry,
)

from ..opentelemetry.observers.execution_observer import observe_execution
from ..util import SERVER_HANDLE_TAG, compressed_namedtuple_upload_file, is_isolated_run
from ..version import __version__
from .instrumentation.constants import DAGSTER_CLOUD_AGENT_METRIC_PREFIX
from .instrumentation.run_launch import extract_run_attributes
from .instrumentation.schedule import inspect_schedule_result
from .instrumentation.sensor import inspect_sensor_result
from .queries import (
    ADD_AGENT_HEARTBEATS_MUTATION,
    DEPLOYMENTS_QUERY,
    GET_USER_CLOUD_REQUESTS_QUERY,
    WORKSPACE_ENTRIES_QUERY,
)

if TYPE_CHECKING:
    import datetime

CHECK_WORKSPACE_INTERVAL_SECONDS = 5

# Interval at which Agent heartbeats are sent to the host cloud
AGENT_HEARTBEAT_INTERVAL_SECONDS = 30

# Interval at which to record utilization metrics
AGENT_UTILIZATION_METRICS_INTERVAL_SECONDS = 60

AGENT_HEARTBEAT_ERROR_LIMIT = 25  # Send at most 25 errors

DEFAULT_PENDING_REQUESTS_LIMIT = 100

SLEEP_INTERVAL_SECONDS = float(os.getenv("DAGSTER_CLOUD_AGENT_SLEEP_INTERVAL_SECONDS", "0.5"))


def UPLOAD_API_RESPONSE_BATCHING_ENABLED():
    return os.getenv("DAGSTER_CLOUD_AGENT_UPLOAD_API_RESPONSE_BATCHING_ENABLED") == "true"


DEPLOYMENT_INFO_QUERY = """
    query DeploymentInfo {
         deploymentInfo {
             deploymentName
         }
     }
"""

AGENT_VERSION_LABEL = "version"
AGENT_DEFAULT_MAX_THREADS_PER_CORE = 10
AGENT_MAX_THREADPOOL_WORKERS = int(
    os.getenv(
        "DAGSTER_CLOUD_AGENT_MAX_THREADPOOL_WORKERS",
        str((os.cpu_count() or 1) * AGENT_DEFAULT_MAX_THREADS_PER_CORE),
    )
)

LIVENESS_CHECK_INTERVAL_SECONDS = float(
    os.getenv("DAGSTER_CLOUD_AGENT_LIVENESS_CHECK_INTERVAL", "15.0")
)


class DagsterCloudAgent:
    def __init__(
        self,
        instance: DagsterCloudAgentInstance,
        pending_requests_limit: int = DEFAULT_PENDING_REQUESTS_LIMIT,
    ):
        self._logger = logging.getLogger("dagster_cloud.agent")
        self._instance: DagsterCloudAgentInstance = instance

        self._batcher: DefaultDict[
            str, Batcher[Tuple[str, DagsterCloudUploadApiResponse], None]
        ] = defaultdict(self._batcher_factory)

        if self._logger.isEnabledFor(logging.DEBUG):
            self._logger.info("Starting Dagster Cloud agent with debug logging...")
        else:
            self._logger.info("Starting Dagster Cloud agent...")

        self._exit_stack = ExitStack()
        self._iteration = 0

        self._executor = self._exit_stack.enter_context(
            FuturesAwareThreadPoolExecutor(
                max_workers=AGENT_MAX_THREADPOOL_WORKERS,
                thread_name_prefix="dagster_cloud_agent_worker",
            )
        )
        self._request_ids_to_futures: Dict[str, Future] = {}
        self._utilization_metrics = init_optional_typeddict(AgentUtilizationMetrics)

        self._last_heartbeat_time: Optional[datetime.datetime] = None

        self._last_workspace_check_time = None

        self._errors: deque = deque(
            maxlen=AGENT_HEARTBEAT_ERROR_LIMIT
        )  # (SerializableErrorInfo, timestamp) tuples

        self._pending_requests: List[Dict[str, Any]] = []
        self._locations_with_pending_requests: Set[Tuple[str, str, bool]] = set()
        self._ready_requests: List[Dict[str, Any]] = []

        self._location_query_times: Dict[Tuple[str, str, bool], float] = {}
        self._pending_requests_limit = check.int_param(
            pending_requests_limit, "pending_requests_limit"
        )
        self._active_deployments: Set[Tuple[str, bool]] = (  # deployment_name, is_branch_deployment
            set()
        )

        self._last_liveness_check_time = None

    def __enter__(self):
        return self

    def __exit__(self, _exception_type, _exception_value, _traceback):
        self._exit_stack.close()

    def _batcher_factory(
        self,
    ) -> Batcher[Tuple[str, DagsterCloudUploadApiResponse], None]:
        return Batcher(
            "upload_api_response",
            self._batch_upload_api_response,
            max_wait_ms=50,
            max_batch_size=32,
        )

    @property
    def _active_deployment_names(self):
        return [deployment[0] for deployment in self._active_deployments]

    @property
    def _active_full_deployment_names(self):
        return [deployment[0] for deployment in self._active_deployments if not deployment[1]]

    def _check_initial_deployment_names(self):
        if self._instance.deployment_names:
            result = self._instance.organization_scoped_graphql_client().execute(
                DEPLOYMENTS_QUERY,
                variable_values={"deploymentNames": self._instance.deployment_names},
            )
            deployments = result["data"]["deployments"]
            existing_deployment_names = {deployment["deploymentName"] for deployment in deployments}
            requested_deployment_names = set(self._instance.deployment_names)
            missing_deployment_names = requested_deployment_names.difference(
                existing_deployment_names
            )

            if missing_deployment_names:
                deployment_str = f"deployment{'s' if len(missing_deployment_names) > 1 else ''} {', '.join(missing_deployment_names)}"
                raise Exception(
                    f"Agent is configured to serve an invalid {deployment_str}. Check your"
                    " agent configuration to make sure it is serving the correct deployment.",
                )

    def _update_agent_resource_limits(
        self, user_code_launcher: DagsterCloudUserCodeLauncher
    ) -> None:
        # The agent should have environment variables defining its resource requests and limits.
        # However, the agent may be running in a container with resource limits that are different
        # For example, on k8s there are ways to effect change on the cpu limit, like mutating admission webhooks.
        # Since the effective cgroup limits precede actual hosts resources when it comes to actual behavior for
        # throttling and oom kills, we attempt to obtain these and fallback on the environment variables.
        container_utilization_metrics = self._utilization_metrics.get("container_utilization", {})
        memory_limit = container_utilization_metrics.get("memory_limit")
        if not memory_limit:
            memory_limit = os.getenv("DAGSTER_CLOUD_AGENT_MEMORY_LIMIT")
            self._logger.info(
                "Cannot obtain cgroup memory limit, using environment value: "
                f"DAGSTER_CLOUD_AGENT_MEMORY_LIMIT={memory_limit}"
            )

        cpu_cfs_period_us = container_utilization_metrics.get("cpu_cfs_period_us")
        cpu_cfs_quota_us = container_utilization_metrics.get("cpu_cfs_quota_us")

        cpu_limit = None
        if cpu_cfs_quota_us and cpu_cfs_period_us:
            cpu_limit = (
                1000.0 * cpu_cfs_quota_us
            ) / cpu_cfs_period_us  # cpu_limit expressed in milliseconds of cpu

        if not cpu_limit:
            cpu_limit = os.getenv("DAGSTER_CLOUD_AGENT_CPU_LIMIT")
            self._logger.info(
                "Cannot obtain CPU CFS values, using environment value: "
                f"DAGSTER_CLOUD_AGENT_CPU_LIMIT={cpu_limit}"
            )

        if not user_code_launcher.user_code_deployment_type.supports_utilization_metrics:
            self._logger.info(
                (
                    f"Cannot interpret resource limits for agent type {user_code_launcher.user_code_deployment_type.value}."
                    "Skipping utilization metrics retrieval."
                )
            )
            return

        limits = {
            "cpu_limit": cpu_limit,
            "memory_limit": memory_limit,
        }

        cpu_request = os.getenv("DAGSTER_CLOUD_AGENT_CPU_REQUEST")
        memory_request = os.getenv("DAGSTER_CLOUD_AGENT_MEMORY_REQUEST")
        if cpu_request:
            limits["cpu_request"] = cpu_request
        if memory_request:
            limits["memory_request"] = memory_request

        self._utilization_metrics["resource_limits"][
            user_code_launcher.user_code_deployment_type.value
        ] = limits  # type: ignore

    def _update_utilization_metrics(self, user_code_launcher: DagsterCloudUserCodeLauncher):
        container_utilization_metrics = retrieve_containerized_utilization_metrics(
            logger=self._logger,
            previous_measurement_timestamp=self._utilization_metrics["container_utilization"][
                "measurement_timestamp"
            ],
            previous_cpu_usage=self._utilization_metrics["container_utilization"]["cpu_usage"],
        )
        self._utilization_metrics["request_utilization"].update(
            self._executor.get_current_utilization_metrics()
        )
        self._utilization_metrics["container_utilization"].update(container_utilization_metrics)

        self._update_agent_resource_limits(user_code_launcher)

        self._logger.info(f"Current utilization metrics: {self._utilization_metrics}")

    def run_loop(
        self,
        user_code_launcher,
        agent_uuid,
    ):
        heartbeat_interval_seconds = AGENT_HEARTBEAT_INTERVAL_SECONDS

        if (
            not self._instance.includes_branch_deployments
            and not self._instance.deployment_names
            and not self._instance.include_all_serverless_deployments
        ):
            self._logger.info(
                "Deployment name was not set - checking to see if it can be fetched from the"
                " server..."
            )
            # Fetch the deployment name from the server if it isn't set (only true
            # for old agents, and only will work if there's a single deployment in the org)
            result = self._instance.graphql_client.execute(DEPLOYMENT_INFO_QUERY)
            deployment_name = result["data"]["deploymentInfo"]["deploymentName"]
            self._instance = self._exit_stack.enter_context(
                DagsterInstance.from_ref(self._instance.ref_for_deployment(deployment_name))  # type: ignore  # (instance subclass)
            )

        self._check_initial_deployment_names()

        self._check_update_workspace(
            user_code_launcher,
            upload_all=user_code_launcher.upload_snapshots_on_startup,
        )

        self._logger.info(
            f"Will start polling for requests from {self._instance.dagster_cloud_url} once user code has"
            " been loaded."
        )

        heartbeat_error_size_limit = int(os.getenv("DAGSTER_CLOUD_AGENT_ERROR_SIZE_LIMIT", "10000"))

        while True:
            try:
                for error in self.run_iteration(user_code_launcher):
                    if error:
                        self._logger.error(str(error))
                        self._errors.appendleft(
                            (
                                truncate_serialized_error(
                                    error, heartbeat_error_size_limit, max_depth=3
                                ),
                                get_current_datetime(),
                            )
                        )
            except Exception:
                error_info = serializable_error_info_from_exc_info(sys.exc_info())
                self._logger.error(f"Caught error:\n{error_info}")
                self._errors.appendleft(
                    (
                        truncate_serialized_error(
                            error_info, heartbeat_error_size_limit, max_depth=3
                        ),
                        get_current_datetime(),
                    )
                )

            # Check for any received interrupts
            with raise_interrupts_as(KeyboardInterrupt):
                pass

            if user_code_launcher.ready_to_serve_requests:
                try:
                    self._check_add_heartbeat(agent_uuid, heartbeat_interval_seconds)
                except Exception:
                    self._logger.exception("Failed to add heartbeat")

            # Check for any received interrupts
            with raise_interrupts_as(KeyboardInterrupt):
                pass

            try:
                self._check_update_workspace(user_code_launcher, upload_all=False)

            except Exception:
                self._logger.error(
                    f"Failed to check for workspace updates: \n{serializable_error_info_from_exc_info(sys.exc_info())}"
                )

            self._write_liveness_sentinel_if_overdue()

            # Check for any received interrupts
            with raise_interrupts_as(KeyboardInterrupt):
                time.sleep(SLEEP_INTERVAL_SECONDS)

    def _write_liveness_sentinel_if_overdue(self):
        if self._last_liveness_check_time is False:
            return

        now = time.time()
        if self._last_liveness_check_time is None:
            self._logger.info("Starting liveness sentinel")
        elif self._last_liveness_check_time + LIVENESS_CHECK_INTERVAL_SECONDS > now:
            return

        # Write to a sentinel file to indicate that we've finished our initial
        # reconciliation - this is used to indicate that we're ready to
        # serve requests
        try:
            if not os.access("/opt", os.W_OK):
                self._logger.warning("Disabling liveness sentinel - /opt is not writable")
                self._last_liveness_check_time = False
                return
            Path("/opt/liveness_sentinel.txt").touch(exist_ok=True)
            self._last_liveness_check_time = now
        except Exception as e:
            self._logger.error(f"Failed to write liveness sentinel and disabling it: {e}")
            self._last_liveness_check_time = False

    def _check_update_workspace(self, user_code_launcher, upload_all):
        curr_time = get_current_datetime()

        if (
            self._last_workspace_check_time
            and (curr_time - self._last_workspace_check_time).total_seconds()
            < CHECK_WORKSPACE_INTERVAL_SECONDS
        ):
            return

        self._last_workspace_check_time = curr_time
        self._query_for_workspace_updates(user_code_launcher, upload_all=upload_all)

    def _check_add_heartbeat(
        self,
        agent_uuid,
        heartbeat_interval_seconds,
    ):
        curr_time = get_current_datetime()

        if (
            self._last_heartbeat_time
            and (curr_time - self._last_heartbeat_time).total_seconds() < heartbeat_interval_seconds
        ):
            return

        errors = [
            TimestampedError(
                timestamp=timestamp.timestamp(),
                error=error,
            )
            for (error, timestamp) in self._errors
            if timestamp.timestamp() > curr_time.timestamp() - 60 * 60 * 24
        ]

        run_worker_statuses_dict = self._instance.user_code_launcher.get_cloud_run_worker_statuses(
            self._active_deployment_names
        )

        code_server_heartbeats_dict = self._instance.user_code_launcher.get_grpc_server_heartbeats()

        agent_image_tag = os.getenv("DAGSTER_CLOUD_AGENT_IMAGE_TAG")
        if self._instance.user_code_launcher.agent_metrics_enabled:
            num_running_requests = self._utilization_metrics["request_utilization"][
                "num_running_requests"
            ]
            max_concurrent_requests = self._utilization_metrics["request_utilization"][
                "max_concurrent_requests"
            ]
            self._logger.info(
                f"Current agent threadpool utilization: {num_running_requests}/{max_concurrent_requests} threads"
            )

        self._last_heartbeat_time = curr_time

        heartbeats = {
            deployment_name: AgentHeartbeat(
                timestamp=curr_time.timestamp(),
                agent_id=agent_uuid,
                agent_label=self._instance.dagster_cloud_api_agent_label,
                agent_type=(
                    type(self._instance.user_code_launcher).__name__
                    if self._instance.user_code_launcher
                    else None
                ),
                metadata=merge_dicts(
                    {AGENT_VERSION_LABEL: __version__},
                    {"image_tag": agent_image_tag} if agent_image_tag else {},
                    {
                        "utilization_metrics": self._utilization_metrics
                        if self._instance.user_code_launcher.agent_metrics_enabled
                        else {}
                    },
                ),
                run_worker_statuses=run_worker_statuses_dict[deployment_name],
                code_server_heartbeats=code_server_heartbeats_dict.get(deployment_name, []),
                agent_queues_config=self._instance.agent_queues_config,
            )
            for deployment_name in self._active_deployment_names
        }

        serialized_agent_heartbeats = [
            {
                "deploymentName": deployment_name,
                "serializedAgentHeartbeat": serialize_value(heartbeat),
            }
            for deployment_name, heartbeat in heartbeats.items()
        ]

        serialized_errors = [serialize_value(error) for error in errors]
        try:
            self._instance.organization_scoped_graphql_client().execute(
                ADD_AGENT_HEARTBEATS_MUTATION,
                variable_values={
                    "serializedAgentHeartbeats": serialized_agent_heartbeats,
                    "serializedErrors": serialized_errors,
                },
                idempotent_mutation=True,
            )
        except DagsterCloudHTTPError as e:
            if e.response.status_code == 413:
                heartbeats_size = sum([len(heartbeat) for heartbeat in serialized_agent_heartbeats])
                errors_size = sum([len(error) for error in serialized_errors])

                error_message = f"Request with {heartbeats_size} heartbeat bytes and {errors_size} error bytes was too large, submitting a smaller heartbeat without messages and stack traces."
                self._logger.exception(error_message)

                serialized_agent_heartbeats = [
                    {
                        "deploymentName": deployment_name,
                        "serializedAgentHeartbeat": serialize_value(
                            heartbeat.without_messages_and_errors()
                        ),
                    }
                    for deployment_name, heartbeat in heartbeats.items()
                ]

                self._instance.organization_scoped_graphql_client().execute(
                    ADD_AGENT_HEARTBEATS_MUTATION,
                    variable_values={
                        "serializedAgentHeartbeats": serialized_agent_heartbeats,
                        "serializedErrors": [
                            serialize_value(
                                TimestampedError(
                                    timestamp=curr_time.timestamp(),
                                    error=SerializableErrorInfo(
                                        error_message,
                                        stack=[],
                                        cls_name=None,
                                    ),
                                )
                            )
                        ],
                    },
                    idempotent_mutation=True,
                )
            else:
                raise

    @property
    def executor(self) -> ThreadPoolExecutor:
        return self._executor

    @property
    def request_ids_to_futures(self) -> Dict[str, Future]:
        return self._request_ids_to_futures

    def _upload_outdated_workspace_entries(
        self,
        deployment_name: str,
        is_branch_deployment: bool,
        user_code_launcher: DagsterCloudUserCodeLauncher,
    ):
        result = self._instance.graphql_client_for_deployment(deployment_name).execute(
            WORKSPACE_ENTRIES_QUERY,
            variable_values={
                "deploymentNames": [deployment_name],
                "includeAllServerlessDeployments": False,
                "agentQueues": self._instance.agent_queues_config.queues,
            },
        )
        entries = result["data"]["deployments"][0]["workspaceEntries"]
        now = time.time()

        upload_metadata = {}

        for entry in entries:
            location_name = entry["locationName"]
            code_location_deploy_data = deserialize_value(
                entry["serializedDeploymentMetadata"], CodeLocationDeployData
            )
            if entry["hasOutdatedData"]:
                # Spin up a server for this location and upload its metadata to Cloud
                # (Bump the TTL counter as well to leave the server up)
                self._location_query_times[
                    (deployment_name, location_name, is_branch_deployment)
                ] = now
                upload_metadata[(deployment_name, location_name)] = UserCodeLauncherEntry(
                    code_location_deploy_data=code_location_deploy_data,
                    update_timestamp=float(entry["metadataTimestamp"]),
                )

        user_code_launcher.add_upload_metadata(upload_metadata)

    def _has_ttl(self, user_code_launcher, is_branch_deployment):
        # branch deployments always have TTLs, other deployments only if you asked for it specifically
        return is_branch_deployment or user_code_launcher.server_ttl_enabled_for_full_deployments

    def _get_ttl_seconds(self, is_branch_deployment):
        return (
            self._instance.user_code_launcher.branch_deployment_ttl_seconds
            if is_branch_deployment
            else self._instance.user_code_launcher.full_deployment_ttl_seconds
        )

    def _get_locations_with_ttl_to_query(self, user_code_launcher) -> List[Tuple[str, str]]:
        now = time.time()

        # For the deployments with TTLs, decide which locations to consider
        # Include the location if:
        # - a) There's a pending request in the queue for it
        # - b) It's TTL hasn't expired since the last time somebody asked for it
        # Always include locations in a), and add locations from b) until you hit a limit
        location_candidates: Set[Tuple[str, str, float]] = {
            (deployment, location, -1.0)  # Score below 0 so that they're at the front of the list
            for deployment, location, is_branch_deployment in self._locations_with_pending_requests
            if self._has_ttl(user_code_launcher, is_branch_deployment)
        }

        num_locations_to_query = self._instance.user_code_launcher.server_ttl_max_servers

        if len(location_candidates) > num_locations_to_query:
            self._logger.warning(
                f"Temporarily keeping {len(location_candidates)} servers with pending requests "
                f"running, which is more than the configured {num_locations_to_query} servers."
            )
            return [(deployment, location) for deployment, location, _score in location_candidates]

        for location_entry, query_time in self._location_query_times.items():
            if location_entry in self._locations_with_pending_requests:
                # Skip locations that are already in location_candidates due to having
                # pending requests
                continue

            deployment_name, location, is_branch_deployment = location_entry

            if not self._has_ttl(user_code_launcher, is_branch_deployment):
                continue

            time_since_last_query = now - query_time

            if time_since_last_query >= self._get_ttl_seconds(is_branch_deployment):
                continue

            location_candidates.add((deployment_name, location, time_since_last_query))

        sorted_results = sorted(location_candidates, key=lambda x: x[2])

        # sort by time since last query asending and return the first N
        filtered_results = sorted_results[:num_locations_to_query]

        num_left_out = len(sorted_results) - num_locations_to_query

        if num_left_out > 0:
            self._logger.warning(
                f"{len(sorted_results)} locations have been queried within TTL, but "
                f"filtering out {num_left_out} to stay within {num_locations_to_query}"
            )

        return [(deployment, location) for deployment, location, _score in filtered_results]

    def _query_for_workspace_updates(
        self,
        user_code_launcher: DagsterCloudUserCodeLauncher,
        upload_all: bool,
    ):
        locations_with_ttl_to_query = self._get_locations_with_ttl_to_query(user_code_launcher)

        deployments_to_query = {key[0] for key in locations_with_ttl_to_query}

        if locations_with_ttl_to_query:
            locations_str = ", ".join(
                f"{deployment}:{location}" for deployment, location in locations_with_ttl_to_query
            )
            self._logger.debug(f"Querying for the following locations with TTL: {locations_str}")

        # If you have specified a non-branch deployment and no TTL, always consider it
        if self._instance.deployment_names:
            deployments_to_query = deployments_to_query.union(set(self._instance.deployment_names))

        # Create mapping of
        # - location name => deployment metadata
        deployment_map: Dict[Tuple[str, str], UserCodeLauncherEntry] = {}
        all_locations: Set[Tuple[str, str]] = set()

        self._active_deployments = set()

        if deployments_to_query or self._instance.include_all_serverless_deployments:
            result = self._instance.organization_scoped_graphql_client().execute(
                WORKSPACE_ENTRIES_QUERY,
                variable_values={
                    "deploymentNames": list(deployments_to_query),
                    "includeAllServerlessDeployments": self._instance.include_all_serverless_deployments,
                    "agentQueues": self._instance.agent_queues_config.queues,
                },
            )

            for deployment_result in result["data"]["deployments"]:
                deployment_name = deployment_result["deploymentName"]
                is_branch_deployment = deployment_result["isBranchDeployment"]

                self._active_deployments.add((deployment_name, is_branch_deployment))

                entries = deployment_result["workspaceEntries"]

                for entry in entries:
                    location_name = entry["locationName"]

                    location_key = (deployment_name, location_name)

                    all_locations.add(location_key)
                    code_location_deploy_data = deserialize_value(
                        entry["serializedDeploymentMetadata"], CodeLocationDeployData
                    )

                    # The GraphQL can return a mix of deployments with TTLs (for example, all
                    # branch deployments, and also full deployments if you have that configured)
                    # and locations without TTLs (for example, full deployments). Deployments
                    # without TTLs always include all of their locations. Deployments with TTLs
                    # only include the locations within locations_with_ttl_to_query.
                    if not self._has_ttl(
                        user_code_launcher, is_branch_deployment
                    ) or location_key in cast(Set[Tuple[str, str]], locations_with_ttl_to_query):
                        deployment_map[location_key] = UserCodeLauncherEntry(
                            code_location_deploy_data=code_location_deploy_data,
                            update_timestamp=float(entry["metadataTimestamp"]),
                        )

        if len(deployment_map):
            update_str = ", ".join(
                f"{deployment}:{location}" for deployment, location in deployment_map.keys()
            )
            self._logger.debug(f"Reconciling with the following locations: {update_str}")
        else:
            self._logger.debug("Reconciling with no locations")

        if upload_all:
            user_code_launcher.add_upload_metadata(deployment_map)
        else:
            user_code_launcher.update_grpc_metadata(deployment_map)

        # Tell run worker monitoring which deployments it should care about
        user_code_launcher.update_run_worker_monitoring_deployments(self._active_deployment_names)

        # In the rare event that there are pending requests that are no longer in the workspace at
        # all (if, say, a location is removed while requests are enqueued), they should be forcibly
        # moved to ready so that they don't stay pending forever - callsites will get an error
        # about the location not existing, but that's preferable to slowly timing out
        pending_requests_copy = self._pending_requests.copy()
        self._pending_requests = []
        for json_request in pending_requests_copy:
            location_name = self._get_location_from_request(json_request)
            deployment_name = json_request["deploymentName"]
            if (deployment_name, location_name) not in all_locations:
                self._ready_requests.append(json_request)
            else:
                self._pending_requests.append(json_request)

    def _get_grpc_client(
        self,
        user_code_launcher: DagsterCloudUserCodeLauncher,
        deployment_name: str,
        location_name: str,
    ) -> DagsterGrpcClient:
        endpoint = user_code_launcher.get_grpc_endpoint(deployment_name, location_name)
        return endpoint.create_client()

    def _get_location_origin_from_request(
        self,
        request: DagsterCloudApiRequest,
    ) -> Optional[CodeLocationOrigin]:
        """Derive the location from the specific argument passed in to a dagster_cloud_api call."""
        api_name = request.request_api
        if api_name in {
            DagsterCloudApi.GET_EXTERNAL_EXECUTION_PLAN,
            DagsterCloudApi.GET_SUBSET_EXTERNAL_PIPELINE_RESULT,
        }:
            external_pipeline_origin = request.request_args.job_origin
            return external_pipeline_origin.repository_origin.code_location_origin
        elif api_name in {
            DagsterCloudApi.GET_EXTERNAL_PARTITION_CONFIG,
            DagsterCloudApi.GET_EXTERNAL_PARTITION_TAGS,
            DagsterCloudApi.GET_EXTERNAL_PARTITION_NAMES,
            DagsterCloudApi.GET_EXTERNAL_PARTITION_SET_EXECUTION_PARAM_DATA,
            DagsterCloudApi.GET_EXTERNAL_SCHEDULE_EXECUTION_DATA,
            DagsterCloudApi.GET_EXTERNAL_SENSOR_EXECUTION_DATA,
        }:
            return request.request_args.repository_origin.code_location_origin
        elif api_name == DagsterCloudApi.GET_EXTERNAL_NOTEBOOK_DATA:
            return request.request_args.code_location_origin
        elif api_name == DagsterCloudApi.PING_LOCATION:
            return RegisteredCodeLocationOrigin(request.request_args.location_name)
        else:
            return None

    @cached_method
    def _get_user_code_instance_ref(self, deployment_name: str):
        return get_instance_ref_for_user_code(self._instance.ref_for_deployment(deployment_name))

    def _handle_api_request(
        self,
        request: DagsterCloudApiRequest,
        deployment_name: str,
        is_branch_deployment: bool,
        user_code_launcher: DagsterCloudUserCodeLauncher,
    ) -> Union[DagsterCloudApiSuccess, DagsterCloudApiGrpcResponse]:
        api_name = request.request_api

        code_location_origin = self._get_location_origin_from_request(request)
        location_name = code_location_origin.location_name if code_location_origin else None

        if api_name == DagsterCloudApi.PING_LOCATION:
            # Do nothing - this request only exists to bump TTL for the location
            return DagsterCloudApiSuccess()
        elif api_name == DagsterCloudApi.CHECK_FOR_WORKSPACE_UPDATES:
            # Dagster Cloud has requested that we upload new metadata for any out of date locations in
            # the workspace
            self._upload_outdated_workspace_entries(
                deployment_name, is_branch_deployment, user_code_launcher
            )
            return DagsterCloudApiSuccess()
        elif api_name == DagsterCloudApi.GET_EXTERNAL_EXECUTION_PLAN:
            client = self._get_grpc_client(
                user_code_launcher, deployment_name, cast(str, location_name)
            )
            serialized_snapshot_or_error = client.execution_plan_snapshot(
                execution_plan_snapshot_args=request.request_args._replace(
                    instance_ref=self._get_user_code_instance_ref(deployment_name)
                )
            )
            return DagsterCloudApiGrpcResponse(
                serialized_response_or_error=serialized_snapshot_or_error
            )

        elif api_name == DagsterCloudApi.GET_SUBSET_EXTERNAL_PIPELINE_RESULT:
            client = self._get_grpc_client(
                user_code_launcher, deployment_name, cast(str, location_name)
            )

            serialized_subset_result_or_error = client.external_pipeline_subset(
                pipeline_subset_snapshot_args=request.request_args
            )

            return DagsterCloudApiGrpcResponse(
                serialized_response_or_error=serialized_subset_result_or_error
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_PARTITION_CONFIG:
            client = self._get_grpc_client(
                user_code_launcher, deployment_name, cast(str, location_name)
            )
            serialized_partition_config_or_error = client.external_partition_config(
                partition_args=request.request_args,
            )
            return DagsterCloudApiGrpcResponse(
                serialized_response_or_error=serialized_partition_config_or_error
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_PARTITION_TAGS:
            client = self._get_grpc_client(
                user_code_launcher, deployment_name, cast(str, location_name)
            )
            serialized_partition_tags_or_error = client.external_partition_tags(
                partition_args=request.request_args,
            )
            return DagsterCloudApiGrpcResponse(
                serialized_response_or_error=serialized_partition_tags_or_error
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_PARTITION_NAMES:
            client = self._get_grpc_client(
                user_code_launcher, deployment_name, cast(str, location_name)
            )
            serialized_partition_names_or_error = client.external_partition_names(
                partition_names_args=request.request_args,
            )
            return DagsterCloudApiGrpcResponse(
                serialized_response_or_error=serialized_partition_names_or_error
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_PARTITION_SET_EXECUTION_PARAM_DATA:
            client = self._get_grpc_client(
                user_code_launcher, deployment_name, cast(str, location_name)
            )
            serialized_partition_execution_params_or_error = (
                client.external_partition_set_execution_params(
                    partition_set_execution_param_args=request.request_args
                )
            )
            return DagsterCloudApiGrpcResponse(
                serialized_response_or_error=serialized_partition_execution_params_or_error
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_SCHEDULE_EXECUTION_DATA:
            client = self._get_grpc_client(
                user_code_launcher, deployment_name, cast(str, location_name)
            )

            args = request.request_args._replace(
                instance_ref=self._get_user_code_instance_ref(deployment_name)
            )

            schedule_attributes = {
                "schedule": args.schedule_name,
                "repository": args.repository_origin.repository_name,
                "location": args.repository_origin.code_location_origin.location_name,
                "deployment": deployment_name,
            }

            with observe_execution(
                opentelemetry=self._instance.opentelemetry,
                event_key=f"{DAGSTER_CLOUD_AGENT_METRIC_PREFIX}.schedule.evaluation",
                short_description="schedule evaluation requests",
                attributes=schedule_attributes,
                result_evaluator_callback=inspect_schedule_result,
            ) as observer:
                serialized_schedule_data_or_error = client.external_schedule_execution(
                    external_schedule_execution_args=args,
                )
                observer.evaluate_result(
                    serialized_data_or_error=serialized_schedule_data_or_error,
                )
                return DagsterCloudApiGrpcResponse(
                    serialized_response_or_error=serialized_schedule_data_or_error
                )

        elif api_name == DagsterCloudApi.GET_EXTERNAL_SENSOR_EXECUTION_DATA:
            client = self._get_grpc_client(
                user_code_launcher, deployment_name, cast(str, location_name)
            )

            args = request.request_args._replace(
                instance_ref=self._get_user_code_instance_ref(deployment_name)
            )

            sensor_attributes = {
                "sensor": args.sensor_name,
                "repository": args.repository_origin.repository_name,
                "location": args.repository_origin.code_location_origin.location_name,
                "deployment": deployment_name,
            }

            with observe_execution(
                opentelemetry=self._instance.opentelemetry,
                event_key=f"{DAGSTER_CLOUD_AGENT_METRIC_PREFIX}.sensor.evaluation",
                short_description="sensor evaluation requests",
                attributes=sensor_attributes,
                result_evaluator_callback=inspect_sensor_result,
            ) as observer:
                serialized_sensor_data_or_error = client.external_sensor_execution(
                    sensor_execution_args=args,
                )
                observer.evaluate_result(serialized_sensor_data_or_error)
                return DagsterCloudApiGrpcResponse(
                    serialized_response_or_error=serialized_sensor_data_or_error
                )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_NOTEBOOK_DATA:
            client = self._get_grpc_client(
                user_code_launcher, deployment_name, cast(str, location_name)
            )
            response = client.external_notebook_data(request.request_args.notebook_path)
            return DagsterCloudApiGrpcResponse(serialized_response_or_error=response.decode())
        elif api_name == DagsterCloudApi.LAUNCH_RUN:
            run = request.request_args.dagster_run

            with DagsterInstance.from_ref(
                self._get_user_code_instance_ref(deployment_name)
            ) as scoped_instance:
                scoped_instance.report_engine_event(
                    f"{self._instance.agent_display_name} is launching run {run.run_id}",
                    run,
                    cls=self.__class__,
                )

                scoped_instance.add_run_tags(
                    run.run_id,
                    merge_dicts(
                        (
                            {"dagster/agent_label": self._instance.dagster_cloud_api_agent_label}
                            if self._instance.dagster_cloud_api_agent_label
                            else {}
                        ),
                        {"dagster/agent_id": self._instance.instance_uuid},
                    ),
                )

                run_attributes = extract_run_attributes(deployment_name, run)
                with observe_execution(
                    opentelemetry=self._instance.opentelemetry,
                    event_key=f"{DAGSTER_CLOUD_AGENT_METRIC_PREFIX}.run.launches",
                    short_description="run execution requests",
                    attributes=run_attributes,
                ) as observer:
                    launcher = scoped_instance.get_run_launcher_for_run(run)  # type: ignore  # (instance subclass)

                    if is_isolated_run(run):
                        launcher.launch_run(LaunchRunContext(dagster_run=run, workspace=None))
                    else:
                        scoped_instance.report_engine_event(
                            f"Launching {run.run_id} without an isolated run environment.",
                            run,
                            cls=self.__class__,
                        )

                        run_location_name = cast(
                            str,
                            run.remote_job_origin.repository_origin.code_location_origin.location_name,
                        )

                        server = user_code_launcher.get_grpc_server(
                            deployment_name, run_location_name
                        )

                        # Record the server handle that we launched it on to for run monitoring
                        scoped_instance.add_run_tags(
                            run.run_id, new_tags={SERVER_HANDLE_TAG: str(server.server_handle)}
                        )

                        launcher.launch_run_from_grpc_client(
                            scoped_instance, run, server.server_endpoint.create_client()
                        )
                    observer.evaluate_result(run=run)

                return DagsterCloudApiSuccess()
        elif api_name == DagsterCloudApi.TERMINATE_RUN:
            # With isolated agents enabled:
            # Run workers now poll for run status. We don't use the run launcher to terminate.
            # Once min agent version is bumped, we can deprecate this command.
            # For backcompat, we use the run launcher to terminate unless the user opts in.
            run = request.request_args.dagster_run

            with DagsterInstance.from_ref(
                self._instance.ref_for_deployment(deployment_name)
            ) as scoped_instance:
                if self._instance.is_using_isolated_agents:
                    scoped_instance.report_engine_event(
                        f"{self._instance.agent_display_name} received request to mark run as canceling",
                        run,
                        cls=self.__class__,
                    )
                    scoped_instance.report_run_canceling(run)
                else:
                    scoped_instance.report_engine_event(
                        f"{self._instance.agent_display_name} received request to terminate run",
                        run,
                        cls=self.__class__,
                    )
                    if is_isolated_run(run):
                        launcher = scoped_instance.get_run_launcher_for_run(run)  # type: ignore  # (instance subclass)
                        launcher.terminate(run.run_id)
                    else:
                        run_location_name = cast(
                            str,
                            run.remote_job_origin.repository_origin.code_location_origin.location_name,
                        )

                        server = user_code_launcher.get_grpc_server(
                            deployment_name, run_location_name
                        )
                        client = server.server_endpoint.create_client()

                        scoped_instance.report_run_canceling(run)
                        client.cancel_execution(CancelExecutionRequest(run_id=run.run_id))

            return DagsterCloudApiSuccess()

        else:
            check.assert_never(api_name)
            raise Exception(f"Unexpected dagster cloud api call {api_name}")

    def _process_api_request(
        self,
        json_request: Dict,
        user_code_launcher: DagsterCloudUserCodeLauncher,
        submitted_to_executor_timestamp: float,
    ) -> Optional[SerializableErrorInfo]:
        thread_start_run_timestamp = get_current_timestamp()
        api_result: Optional[DagsterCloudApiResponse] = None
        error_info: Optional[SerializableErrorInfo] = None

        request_id = json_request["requestId"]
        request_api = json_request["requestApi"]
        request_body = json_request["requestBody"]
        deployment_name = json_request["deploymentName"]
        is_branch_deployment = json_request["isBranchDeployment"]

        request: Union[str, DagsterCloudApiRequest] = DagsterCloudApiRequest.format_request(
            request_id, request_api
        )

        if request_api not in DagsterCloudApi.__members__:
            api_result = DagsterCloudApiUnknownCommandResponse(request_api=request_api)
            self._logger.warning(
                f"Ignoring request {json_request}: Unknown command. This is likely due to running an "
                "older version of the agent."
            )
        else:
            try:
                request = deserialize_value(request_body, DagsterCloudApiRequest)
                self._logger.info(f"Received request {request}.")
                api_result = self._handle_api_request(
                    request, deployment_name, is_branch_deployment, user_code_launcher
                )
            except Exception:
                error_info = serializable_error_info_from_exc_info(sys.exc_info())
                api_result = DagsterCloudApiErrorResponse(error_infos=[error_info])

                self._logger.error(f"Error serving request {json_request}: {error_info}")

        self._logger.info(f"Finished processing request {request}.")

        thread_finished_request_time = get_current_timestamp()
        thread_telemetry = DagsterCloudApiThreadTelemetry(
            submitted_to_executor_timestamp=submitted_to_executor_timestamp,
            thread_start_run_timestamp=thread_start_run_timestamp,
            thread_end_handle_api_request_timestamp=thread_finished_request_time,
        )

        api_result = api_result.with_thread_telemetry(thread_telemetry)

        assert api_result

        upload_response = DagsterCloudUploadApiResponse(
            request_id=request_id,
            request_api=request_api,
            response=api_result,
        )

        self._logger.info(f"Uploading response for request {request}.")

        if UPLOAD_API_RESPONSE_BATCHING_ENABLED():
            self._batcher[deployment_name].submit((deployment_name, upload_response))
        else:
            upload_api_response(self._instance, deployment_name, upload_response)

        self._logger.info(f"Finished uploading response for request {request}.")

        return error_info

    def _batch_upload_api_response(
        self, upload_response_batch: List[Tuple[str, DagsterCloudUploadApiResponse]]
    ) -> List[None]:
        deployment_names = set(deployment_name for deployment_name, _ in upload_response_batch)
        assert len(deployment_names) == 1
        batch_upload_api_response(
            self._instance,
            next(iter(deployment_names)),
            [resp for _, resp in upload_response_batch],
        )
        return [None for _ in upload_response_batch]

    def _get_location_from_request(self, json_request: Dict[str, Any]) -> Optional[str]:
        request_api = json_request["requestApi"]
        request_body = json_request["requestBody"]
        if request_api not in DagsterCloudApi.__members__:
            return None

        request = deserialize_value(request_body, DagsterCloudApiRequest)
        location_origin = self._get_location_origin_from_request(request)
        if not location_origin:
            return None

        return location_origin.location_name

    def run_iteration(
        self, user_code_launcher: DagsterCloudUserCodeLauncher
    ) -> Iterator[Optional[SerializableErrorInfo]]:
        if not user_code_launcher.ready_to_serve_requests:
            return

        num_pending_requests = len(self._pending_requests)

        if num_pending_requests < self._pending_requests_limit:
            # limit (implicit default 10) applied separately for requests to branch deployments, and for each full deployment
            result = self._instance.organization_scoped_graphql_client().execute(
                GET_USER_CLOUD_REQUESTS_QUERY,
                {
                    "forBranchDeployments": self._instance.includes_branch_deployments,
                    "forFullDeployments": self._active_full_deployment_names,
                    "agentQueues": self._instance.agent_queues_config.queues,
                },
            )
            json_requests = result["data"]["userCloudAgent"]["popUserCloudAgentRequests"]

            self._logger.debug(
                f"Iteration #{self._iteration}: Adding {len(json_requests)} requests to be"
                f" processed. Currently {num_pending_requests} waiting for server to be ready"
            )
            self._pending_requests.extend(json_requests)

        else:
            self._logger.warning(
                f"Iteration #{self._iteration}: Waiting to pull requests from the queue since there are"
                f" already {len(self._pending_requests)} in the queue"
            )

        invalid_requests = []
        self._locations_with_pending_requests = set()

        # Determine which pending requests are now ready (their locations have been loaded, or the
        # request does not correspond to a particular location)
        for json_request in self._pending_requests:
            deployment_name = json_request["deploymentName"]
            is_branch_deployment = json_request["isBranchDeployment"]
            location_name = self._get_location_from_request(json_request)
            if location_name:
                self._location_query_times[
                    (deployment_name, location_name, is_branch_deployment)
                ] = time.time()

                if not user_code_launcher.has_grpc_endpoint(deployment_name, location_name):
                    # Next completed periodic workspace update will make the location up to date
                    # - keep this in the queue until then
                    invalid_requests.append(json_request)
                    self._locations_with_pending_requests.add(
                        (deployment_name, location_name, is_branch_deployment)
                    )
                    continue

            self._ready_requests.append(json_request)

        # Any invalid requests go back in the pending queue - the next workspace update will
        # ensure that the usercodelauncher spins up locations for those requests
        self._pending_requests = invalid_requests

        # send all ready requests to the threadpool
        for json_request in self._ready_requests:
            request_id = json_request["requestId"]
            submitted_to_executor_timestamp = get_current_timestamp()
            future = self._executor.submit(
                self._process_api_request,
                json_request,
                user_code_launcher,
                submitted_to_executor_timestamp,
            )

            self._request_ids_to_futures[request_id] = future

        self._ready_requests = []

        # Process futures that are done
        # Create a shallow copy of the futures dict to modify it while iterating
        for request_id, future in self._request_ids_to_futures.copy().items():
            if future.done():
                response: Optional[SerializableErrorInfo] = None

                try:
                    response = future.result(timeout=0)
                except:
                    response = serializable_error_info_from_exc_info(sys.exc_info())

                # Do not process a request again once we have its result
                del self._request_ids_to_futures[request_id]

                # Yield the error information from the future
                if response:
                    yield response

        if self._instance.user_code_launcher.agent_metrics_enabled and (
            self._utilization_metrics["container_utilization"]["measurement_timestamp"] is None
            or (
                get_current_timestamp()
                - self._utilization_metrics["container_utilization"]["measurement_timestamp"]
                > AGENT_UTILIZATION_METRICS_INTERVAL_SECONDS
            )
        ):
            self._update_utilization_metrics(user_code_launcher)
        self._iteration += 1

        yield None


def batch_upload_api_response(
    instance: DagsterCloudAgentInstance,
    deployment_name: str,
    batch: List[DagsterCloudUploadApiResponse],
):
    with compressed_namedtuple_upload_file(BatchDagsterCloudUploadApiResponse(batch=batch)) as f:
        resp = instance.requests_managed_retries_session.put(
            instance.dagster_cloud_upload_api_response_url,
            headers=instance.headers_for_deployment(deployment_name),
            files={"api_response_batch.tmp": f},
            timeout=instance.dagster_cloud_api_timeout,
            proxies=instance.dagster_cloud_api_proxies,
        )
        raise_http_error(resp)


def upload_api_response(
    instance: DagsterCloudAgentInstance,
    deployment_name: str,
    upload_response: DagsterCloudUploadApiResponse,
):
    with compressed_namedtuple_upload_file(upload_response) as f:
        resp = instance.requests_managed_retries_session.put(
            instance.dagster_cloud_upload_api_response_url,
            headers=instance.headers_for_deployment(deployment_name),
            files={"api_response.tmp": f},
            timeout=instance.dagster_cloud_api_timeout,
            proxies=instance.dagster_cloud_api_proxies,
        )
        raise_http_error(resp)



import logging
from threading import Lock
from typing import TYPE_CHECKING, Any, Mapping, Optional

import dagster_cloud.opentelemetry.factories.logs as logs_factory
import dagster_cloud.opentelemetry.factories.metrics as metrics_factory
from dagster_cloud.opentelemetry.metrics.meter import Meter

if TYPE_CHECKING:
    from opentelemetry._logs import LoggerProvider as APILoggerProvider
    from opentelemetry.metrics import MeterProvider as APIMeterProvider
    from opentelemetry.sdk._logs import LogRecordProcessor
    from opentelemetry.sdk._logs.export import LogExporter
    from opentelemetry.sdk.metrics.export import MetricExporter, MetricReader


DEFAULT_SHUTDOWN_TIMEOUT_MILLIS = 5_000


class OpenTelemetryController:
    """The OpenTelemetryController manages the initialization and shutdown of OpenTelemetry components.

    Provided with a configuration, the OpenTelemetryController will initialize the logging and metrics components.
    The python logging will be instrumented with OpenTelemetry and captured and exported.

    The get_meter() method allow to obtain a meter than can create metric instruments.
    See: https://opentelemetry.io/docs/languages/python/instrumentation/#metrics

    Note: As some OpenTelemetry pipeline components run their own thread, it's important to shut down properly.

    Examples:
        .. code-block:: python
            import logging
            import sys
            import time

            from dagster_cloud.opentelemetry.controller import OpenTelemetryController

            otel_config = {
                "enabled": True,
                "logging": {
                    "enabled": True,
                    "exporter": {
                        "type": "ConsoleLogExporter",
                    },
                },
                "metrics": {
                    "enabled": True,
                    "reader": {
                        "type": "PeriodicExportingMetricReader",
                        "params": {
                            # this is the interval at which metrics are captured and exported
                            "export_interval_millis": 800,
                        },
                    },
                    "exporter": {
                        "type": "ConsoleMetricExporter",
                    },
                },
            }


            def main():
                logging.basicConfig(level=logging.INFO, stream=sys.stdout)
                logger = logging.getLogger("my_logger")

                logger.info("OpenTelemetry is not started. This log will not be captured by OpenTelemetry.")
                opentelemetry = OpenTelemetryController(
                    service_name="my_service",
                    instance_id="my_instance",
                    version="0.0.1",
                    config=otel_config,
                    logger=logger,  # this and its children will be instrumented with OpenTelemetry
                )
                logger.info("OpenTelemetry started. This log will be exported by OpenTelemetry.")
                meter = opentelemetry.get_meter(name="item_processor", attributes={"environment": "dev"})
                # the counter should be captured as a metric
                counter = meter.create_counter(
                    name="items_processed", description="The number of items processed.", unit="unit"
                )

                # simulate doing some work, and emitting metrics that will be exported by OpenTelemetry
                counter.add(1)
                time.sleep(1)
                counter.add(1)
                time.sleep(1)

                opentelemetry.dispose()
                logger.info("OpenTelemetry stopped. This log will not be captured by OpenTelemetry.")


            if __name__ == "__main__":
                main()
    """

    def __init__(
        self,
        service_name: str,
        instance_id: str,
        version: str,
        config: Optional[Mapping[str, Any]] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """Initialize the telemetry instance.
        params:
            service_name: str: The name of the service.
            instance_id: str: The unique identifier of the service instance.
            version: str: The version of the service.
            config: dict: The configuration for the telemetry components.
            logger: Logger: The logger to use for telemetry logs. Defaults to the root logger.
        """
        self._shutdown = False

        self._log_exporter: Optional[LogExporter] = None
        self._logger_provider: APILoggerProvider = logs_factory.build_noop_logger_provider()
        self._log_record_processor: Optional[LogRecordProcessor] = None
        self._logging_handler: Optional[logging.Handler] = None

        self._metric_exporter: Optional[MetricExporter] = None
        self._metric_reader: Optional[MetricReader] = None
        self._meter_provider: APIMeterProvider = metrics_factory.build_noop_meter_provider()
        self._meters: dict[str, Meter] = {}

        self._config = config or {}
        self._logger = logger or logging.getLogger()
        self._resource_attributes = {
            "service.name": service_name,
            "service.version": version,
            "service.instance.id": instance_id,
        }
        self._lock = Lock()
        self._initialize()

    def _initialize(self):
        """Initialize the OpenTelemetry components."""
        # an initialized controller can be used in code without having to check if it's
        # enabled as the components will default to NoOp implementations if disabled
        if not self.enabled:
            self._logger.debug(
                "OpenTelemetry is disabled by configuration. Telemetry will not be captured."
            )
            return

        self._initialize_logging()
        self._initialize_metrics()
        # TODO - self._initialize_tracing()

    @property
    def enabled(self):
        """Check if the OpenTelemetry instance is enabled. If not enabled,
        all the components will default to NoOp implementations.
        """
        if not self._config:
            return False

        if not self._config.get("enabled", False):
            return False

        # If no specific opentelemetry components are enabled, then we can effectively disable opentelemetry
        return any(
            [
                self._config.get("metrics", {}).get("enabled", False),
                self._config.get("logging", {}).get("enabled", False),
                self._config.get("tracing", {}).get("enabled", False),
            ]
        )

    @property
    def active(self) -> bool:
        """Check if the OpenTelemetry instance is active.

        Returns: bool: True if the instance is active, otherwise False.
        """
        return not self._shutdown

    @property
    def logging_enabled(self) -> bool:
        """Check if the logging component is enabled."""
        return self.enabled and self._config.get("logging", {}).get("enabled", False)

    @property
    def metrics_enabled(self) -> bool:
        """Check if the metrics component is enabled."""
        return self.enabled and self._config.get("metrics", {}).get("enabled", False)

    @property
    def tracing_enabled(self) -> bool:
        """Check if the tracing component is enabled."""
        return self.enabled and self._config.get("tracing", {}).get("enabled", False)

    def dispose(self):
        """Shutdown the OpenTelemetry components."""
        with self._lock:
            if self._shutdown:
                self._logger.warning("OpenTelemetry: Already shutdown, skipping.")
                return

            # TODO - self._shutdown_tracing()
            self._shutdown_metrics()
            self._shutdown_logging()
            self._shutdown = True

    def get_meter(
        self,
        name: str,
        version: Optional[str] = None,
        schema_url: Optional[str] = None,
        attributes: Optional[Mapping[str, Any]] = None,
    ) -> Meter:
        """Get or create a meter.

        params:
            name: str: The name of the meter.
            version: str: The version of the meter.
            schema_url: str: The schema URL of the meter.
            attributes: dict: The attributes of the meter.

        Returns: Meter: The meter instance.
        """
        if self._shutdown:
            raise RuntimeError("OpenTelemetry: Cannot get meter, already shutdown.")

        with self._lock:
            # guarding with a lock because concurrent calls to get_meter
            # could attempt to create multiple identical meters
            if name not in self._meters:
                meter = self._meter_provider.get_meter(
                    name=name,
                    version=version,
                    schema_url=schema_url,
                    attributes=attributes,
                )
                self._meters[name] = Meter(meter)
            return self._meters[name]

    def _initialize_logging(self):
        logging_config = self._config.get("logging", {})
        if not logging_config.get("enabled"):
            self._logger.info(
                "OpenTelemetry: Logging instrumentation is disabled by configuration."
            )
            return

        self._log_exporter = logs_factory.build_log_exporter(logging_config.get("exporter"))  # pyright: ignore[reportArgumentType]
        self._log_record_processor = logs_factory.build_log_record_processor(
            self._log_exporter, logging_config.get("processor", {})
        )
        self._logger_provider = logs_factory.build_logger_provider(
            self._log_record_processor, self._resource_attributes
        )
        self._logging_handler = logs_factory.build_logging_handler(
            self._logger_provider,  # pyright: ignore[reportArgumentType]
            logging_config.get("handler", {}),  # pyright: ignore[reportArgumentType]
        )
        self._logger.addHandler(self._logging_handler)

    def _initialize_metrics(self):
        metrics_config = self._config.get("metrics", {})
        if not metrics_config.get("enabled"):
            self._logger.info("OpenTelemetry: Metric instrumentation is disabled by configuration.")
            return

        self._metric_exporter = metrics_factory.build_metric_exporter(
            metrics_config.get("exporter", {})
        )

        self._metric_reader = metrics_factory.build_metric_reader(
            self._metric_exporter, metrics_config.get("reader", {})
        )
        self._meter_provider = metrics_factory.build_meter_provider(
            [self._metric_reader], self._resource_attributes
        )

    def _initialize_tracing(self):
        tracing_config = self._config.get("tracing", {})
        if not tracing_config.get("enabled"):
            self._logger.info(
                "OpenTelemetry: Tracing instrumentation is disabled by configuration."
            )
            return
        raise NotImplementedError("Tracing instrumentation is not yet implemented.")

    def _shutdown_logging(self):
        if self._logger and self._logging_handler:
            self._logging_handler.flush()
            self._logger.removeHandler(self._logging_handler)

        # The LoggerProvider will shut down the record processor, which in turns handles the exporter.
        if self._logger_provider:
            # The NoOpLoggerProvider does not have a force_flush or shutdown method.
            if _validate_object_has_method(self._logger_provider, "force_flush"):
                self._logger_provider.force_flush(timeout_millis=DEFAULT_SHUTDOWN_TIMEOUT_MILLIS)  # type: ignore
            if _validate_object_has_method(self._logger_provider, "shutdown"):
                self._logger_provider.shutdown()  # pyright: ignore[reportAttributeAccessIssue]

    def _shutdown_metrics(self):
        # The meter provider will shut down the reader and the PeriodicExportingMetricReader will handle the exporter.
        # Note: the InMemoryMetricReader will not shut down the exporter, it should only be used for testing and
        # does not accept an exporter and one should not exist.
        if self._meter_provider:
            # Duck typing: The base interface and the NoOpMeterProvider do not have a force_flush or shutdown method.
            if _validate_object_has_method(self._meter_provider, "force_flush"):
                self._meter_provider.force_flush(timeout_millis=DEFAULT_SHUTDOWN_TIMEOUT_MILLIS)  # type: ignore
            if _validate_object_has_method(self._meter_provider, "shutdown"):
                self._meter_provider.shutdown()  # type: ignore
        self._meters = {}

    def _shutdown_tracing(self):
        #  TODO - Tracing instrumentation is not yet implemented.
        pass


def _validate_object_has_method(obj, method_name) -> bool:
    """Check if an object has a callable method.

    Some OpenTelemetry components do not have some methods, so we need to check if they exist before calling them.
    """
    return hasattr(obj, method_name) and callable(getattr(obj, method_name))



import os
import sys
from typing import Sequence

from dagster._serdes.ipc import interrupt_ipc_subprocess_pid, open_ipc_subprocess

from . import TaskStatus


def launch_process(args: Sequence[str]) -> int:
    """Launch a process and return the PID."""
    p = open_ipc_subprocess(args)
    pid = p.pid
    return pid


def check_on_process(pid: int) -> TaskStatus:
    # TODO: implement cross platform process check
    if sys.platform == "win32":
        return TaskStatus.NOT_IMPLEMENTED
    else:
        try:
            # Send a no-op. If the process is not running, it will respond with an error
            # https://stackoverflow.com/a/568285/14656695
            os.kill(pid, 0)
        except OSError:
            return TaskStatus.NOT_FOUND
        else:
            return TaskStatus.RUNNING


def kill_process(pid: int):
    interrupt_ipc_subprocess_pid(pid)



from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Callable, Dict, Optional

from dagster import DagsterError
from dagster._time import get_current_timestamp

from ..controller import OpenTelemetryController
from .dagster_exception_handler import extract_dagster_error_attributes

ResultEvaluatorCallback = Callable[..., str]


class ExecutionObserverInstruments:
    """OpenTelemetry instruments used to record generic code block executions."""

    def __init__(
        self, opentelemetry: OpenTelemetryController, metric_base_name: str, description: str
    ):
        self._meter = opentelemetry.get_meter(
            name=metric_base_name,
            attributes={
                "description": description,
            },
        )

        self._executions = self._meter.get_counter(
            name=metric_base_name,
            description=f"Number of {description}",
        )

        self._completions = self._meter.get_counter(
            name=f"{metric_base_name}.completions",
            description=f"Number of {description} completions by status",
        )

        self._exceptions = self._meter.get_counter(
            name=f"{metric_base_name}.exceptions",
            description=f"Number of exceptions caught in {description}",
        )

        self._duration = self._meter.get_histogram(
            name=f"{metric_base_name}.duration",
            description=f"Duration of {description}",
            unit="seconds",
        )

    @property
    def meter(self):
        """Get the OpenTelemetry meter. This is useful for adding custom metrics."""
        return self._meter

    @property
    def executions(self):
        """Get the OpenTelemetry counter for the number of executions of the observed code block."""
        return self._executions

    @property
    def completions(self):
        """Get the OpenTelemetry counter for the number of completions of the observed code block."""
        return self._completions

    @property
    def exceptions(self):
        """Get the OpenTelemetry counter for the number of exceptions caught in the observed code block."""
        return self._exceptions

    @property
    def duration(self):
        """Get the OpenTelemetry gauge for the duration of the observed code block."""
        return self._duration


class ExecutionResultObserver(ABC):
    @abstractmethod
    def evaluate_result(self, *args, **kwargs):
        raise NotImplementedError("Subclasses must implement this method")


class _NoopExecutionResultObserver(ExecutionResultObserver):
    """A no-op observer that does not record any metrics."""

    def evaluate_result(self, *args, **kwargs):
        pass


class _DagsterExecutionResultObserver(ExecutionResultObserver):
    def __init__(
        self,
        instruments: ExecutionObserverInstruments,
        attributes: Dict[str, str],
        result_evaluator_callback: Optional[ResultEvaluatorCallback] = None,
    ):
        self._result_evaluator_callback = result_evaluator_callback
        self._instruments = instruments
        self._attributes = attributes or {}
        self._status = None

    @property
    def meter(self):
        """Get the OpenTelemetry meter. This is useful for adding custom metrics."""
        return self._instruments.meter

    def evaluate_result(self, *args, **kwargs):
        """Evaluate the result of the observed code block.

        If a result evaluator callback is present, it will be called with the arguments passed to
        this method, as well as the attributes and instruments passed to the observer.

        If no result evaluator callback is present, the status will be set to "success".
        """
        if self._result_evaluator_callback:
            self._status = self._result_evaluator_callback(
                *args, attributes=self._attributes, instruments=self._instruments, **kwargs
            )
        else:
            self._status = "success"

    def set_status(self, status: str):
        self._status = status

    def get_status(self) -> Optional[str]:
        return self._status


@contextmanager
def observe_execution(
    opentelemetry: Optional[OpenTelemetryController],
    event_key: str,
    short_description: str,
    result_evaluator_callback: Optional[ResultEvaluatorCallback] = None,
    attributes: Optional[Dict[str, str]] = None,
):
    """This context manager is used to observe the execution of a code block.

    It records the number of executions, completions, exceptions, and the duration of the code block.
    An optional evaluator callback can be provided to evaluate the result of the code block,
    setting a completion status and also optionally implementing other side effects such as
    recording additional metrics or logging based on the results.
    """
    attributes = attributes or {}
    if opentelemetry and opentelemetry.metrics_enabled:
        instruments = ExecutionObserverInstruments(
            opentelemetry=opentelemetry,
            metric_base_name=event_key,
            description=short_description,
        )
        instruments.executions.add(1, attributes)

        start_time = get_current_timestamp()
        observer = _DagsterExecutionResultObserver(
            instruments=instruments,
            attributes=attributes,
            result_evaluator_callback=result_evaluator_callback,
        )

        try:
            yield observer
        except DagsterError as e:
            err_attributes = extract_dagster_error_attributes(e)
            observer.set_status("exception")
            instruments.exceptions.add(1, attributes={**attributes, **err_attributes})
            raise
        except Exception as e:
            err_attributes = {"exception": type(e).__name__, **attributes}
            observer.set_status("exception")
            instruments.exceptions.add(1, err_attributes)
            raise
        finally:
            instruments.completions.add(
                1, attributes={**attributes, "status": observer.get_status() or "unknown"}
            )
            instruments.duration.record(get_current_timestamp() - start_time, attributes)

    else:
        yield _NoopExecutionResultObserver()



from typing import TYPE_CHECKING, Any, Iterable, Optional, Sequence, Set

import dagster._check as check
from dagster._core.definitions.asset_key import EntityKey
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AutomationConditionEvaluationWithRunIds,
)
from dagster._core.definitions.run_request import InstigatorType
from dagster._core.scheduler.instigation import (
    AutoMaterializeAssetEvaluationRecord,
    InstigatorState,
    InstigatorStatus,
    InstigatorTick,
    TickData,
    TickStatus,
)
from dagster._core.storage.schedules.base import ScheduleStorage
from dagster._serdes import (
    ConfigurableClass,
    ConfigurableClassData,
    deserialize_value,
    serialize_value,
)
from typing_extensions import Self

from .queries import (
    ADD_JOB_STATE_MUTATION,
    ALL_STORED_JOB_STATE_QUERY,
    CREATE_JOB_TICK_MUTATION,
    GET_JOB_STATE_QUERY,
    GET_JOB_TICKS_QUERY,
    UPDATE_JOB_STATE_MUTATION,
    UPDATE_JOB_TICK_MUTATION,
)

if TYPE_CHECKING:
    from dagster_cloud.instance import DagsterCloudAgentInstance  # noqa: F401


class GraphQLScheduleStorage(ScheduleStorage["DagsterCloudAgentInstance"], ConfigurableClass):
    def __init__(self, inst_data=None, override_graphql_client=None):
        """Initialize this class directly only for test (using `override_graphql_client`).
        Use the ConfigurableClass machinery to init from instance yaml.
        """
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        self._override_graphql_client = override_graphql_client

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @classmethod
    def from_config_value(cls, inst_data: ConfigurableClassData, config_value: Any) -> Self:
        return cls(inst_data=inst_data)

    @property
    def _graphql_client(self):
        return (
            self._override_graphql_client
            if self._override_graphql_client
            else self._instance.graphql_client
        )

    def _execute_query(self, query, variables=None):
        return self._graphql_client.execute(query, variable_values=variables)

    def wipe(self):
        raise Exception("Not allowed to wipe from user cloud")

    def all_instigator_state(  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self,
        repository_origin_id: Optional[str] = None,
        repository_selector_id: Optional[str] = None,
        instigator_type: Optional[InstigatorType] = None,
        instigator_statuses: Optional[Set[InstigatorStatus]] = None,
    ) -> Iterable[InstigatorState]:
        res = self._execute_query(
            ALL_STORED_JOB_STATE_QUERY,
            variables={
                "repositoryOriginId": repository_origin_id,
                "repositorySelectorId": repository_selector_id,
                "jobType": instigator_type.value if instigator_type else None,
                "statuses": (
                    [status.value for status in instigator_statuses]
                    if instigator_statuses
                    else None
                ),
            },
        )

        return [
            deserialize_value(state, InstigatorState)
            for state in res["data"]["schedules"]["jobStates"]
        ]

    def get_instigator_state(self, origin_id: str, selector_id: str) -> Optional[InstigatorState]:
        res = self._execute_query(
            GET_JOB_STATE_QUERY,
            variables={
                "jobOriginId": origin_id,
                "selectorId": selector_id,
            },
        )

        state = res["data"]["schedules"]["jobState"]
        if state is None:
            return None

        return deserialize_value(state, InstigatorState)

    def add_instigator_state(self, state: InstigatorState):  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self._execute_query(
            ADD_JOB_STATE_MUTATION,
            variables={
                "serializedJobState": serialize_value(
                    check.inst_param(state, "state", InstigatorState)
                )
            },
        )

    def update_instigator_state(self, state: InstigatorState):  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self._execute_query(
            UPDATE_JOB_STATE_MUTATION,
            variables={
                "serializedJobState": serialize_value(
                    check.inst_param(state, "state", InstigatorState)
                )
            },
        )

    def delete_instigator_state(self, origin_id: str, selector_id: str):
        raise NotImplementedError("Not callable from user cloud")

    def get_tick(self, tick_id: int) -> InstigatorTick:
        raise NotImplementedError("Not callable from user cloud")

    def get_ticks(  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self,
        origin_id: str,
        selector_id: str,
        before: Optional[float] = None,
        after: Optional[float] = None,
        limit: Optional[int] = None,
        statuses: Optional[Sequence[TickStatus]] = None,
    ) -> Iterable[InstigatorTick]:
        status_strs = [status.value for status in statuses] if statuses else None
        res = self._execute_query(
            GET_JOB_TICKS_QUERY,
            variables={
                "jobOriginId": origin_id,
                "selectorId": selector_id,
                "before": before,
                "after": after,
                "limit": limit,
                "statuses": status_strs,
            },
        )

        return [
            deserialize_value(tick, InstigatorTick) for tick in res["data"]["schedules"]["jobTicks"]
        ]

    def create_tick(self, tick_data: TickData):
        res = self._execute_query(
            CREATE_JOB_TICK_MUTATION,
            variables={
                "serializedJobTickData": serialize_value(
                    check.inst_param(tick_data, "tick_data", TickData)
                )
            },
        )

        tick_id = res["data"]["schedules"]["createJobTick"]["tickId"]
        return InstigatorTick(tick_id, tick_data)

    def update_tick(self, tick: InstigatorTick):
        self._execute_query(
            UPDATE_JOB_TICK_MUTATION,
            variables={
                "tickId": tick.tick_id,
                "serializedJobTickData": serialize_value(
                    check.inst_param(tick.tick_data, "tick_data", TickData)
                ),
            },
        )
        return tick

    def purge_ticks(
        self,
        origin_id: str,
        selector_id: str,
        before: float,
        tick_statuses: Optional[Sequence[TickStatus]] = None,
    ):
        raise NotImplementedError("Not callable from user cloud")

    def add_auto_materialize_asset_evaluations(
        self,
        evaluation_id: int,
        asset_evaluations: Sequence[AutomationConditionEvaluationWithRunIds],
    ) -> None:
        raise NotImplementedError("Not callable from user cloud")

    def get_auto_materialize_asset_evaluations(
        self, key: EntityKey, limit: int, cursor: Optional[int] = None
    ) -> Sequence[AutoMaterializeAssetEvaluationRecord]:
        raise NotImplementedError("Not callable from user cloud")

    def get_auto_materialize_evaluations_for_evaluation_id(
        self, evaluation_id: int
    ) -> Sequence[AutoMaterializeAssetEvaluationRecord]:
        raise NotImplementedError("Not callable from user cloud")

    def purge_asset_evaluations(self, before: float):
        raise NotImplementedError("Not callable from user cloud")

    def upgrade(self):
        raise NotImplementedError("Not callable from user cloud")

    def optimize_for_dagit(self, statement_timeout: int, pool_recycle: int):
        raise NotImplementedError("Not callable from user cloud")



import warnings
from dataclasses import dataclass
from datetime import date, datetime, timedelta, timezone
from pprint import pprint
from typing import TYPE_CHECKING, Optional, Sequence, Union

from dagster import (
    AssetExecutionContext,
    AssetsDefinition,
    AssetSelection,
    HourlyPartitionsDefinition,
    ScheduleDefinition,
    ScheduleEvaluationContext,
    TimeWindow,
    asset,
    build_schedule_from_partitioned_job,
    define_asset_job,
    fs_io_manager,
    schedule,
)

from ..metrics_utils import put_cost_information
from .dagster_snowflake_insights import get_cost_data_for_hour

if TYPE_CHECKING:
    from dagster_snowflake import SnowflakeConnection

SNOWFLAKE_QUERY_HISTORY_LATENCY_SLA_MINS = 45


@dataclass
class SnowflakeInsightsDefinitions:
    assets: Sequence[AssetsDefinition]
    schedule: ScheduleDefinition


def create_snowflake_insights_asset_and_schedule(
    start_date: Union[datetime, date, str],
    name: Optional[str] = None,
    group_name: Optional[str] = None,
    job_name: str = "snowflake_insights_import",
    dry_run=False,
    allow_partial_partitions=True,
    snowflake_resource_key: str = "snowflake",
    snowflake_usage_latency: int = SNOWFLAKE_QUERY_HISTORY_LATENCY_SLA_MINS,
    partition_end_offset_hrs: int = 1,
    schedule_batch_size_hrs: int = 1,
    submit_to_s3_only: bool = True,  # DEPRECATED
) -> SnowflakeInsightsDefinitions:
    """Generates a pre-defined Dagster asset and schedule that can be used to import Snowflake cost
    data into Dagster Insights.

    The schedule will run hourly, and will query the Snowflake query_history table for all queries
    that ran in the hour starting at the scheduled time. It will then submit the cost data to
    Dagster Insights.

    Args:
        start_date (Union[datetime, str]): The date to start the partitioned schedule on. This should be the date
            that you began to track cost data alongside your dbt runs.
        name (Optional[str]): The name of the asset. Defaults to "snowflake_query_history".
        group_name (Optional[str]): The name of the asset group. Defaults to the default group.
        job_name (str): The name of the job that will be created to run the schedule. Defaults to
            "snowflake_insights_import".
        dry_run (bool): If true, the schedule will print the cost data to stdout instead of
            submitting it to Dagster Insights. Defaults to True.
        snowflake_resource_key (str): The name of the snowflake resource key to use. Defaults to
            "snowflake".
        partition_end_offset_hrs (int): The number of additional hours to wait before querying
            Snowflake for the latest data. This is useful in case the Snowflake query_history table
            is not immediately available. Defaults to 1
        schedule_batch_size_hrs (int): The number of hours of data to process in each schedule
            run. For example, if this is set to 2, the schedule will run every 2 hours and process
            2 hours of data. Defaults to 1.
    """
    # for backcompat, this used to take `date`
    if isinstance(start_date, date):
        start_date = start_date.strftime("%Y-%m-%d-%H:%M")

    if submit_to_s3_only is False:
        warnings.warn(
            "The `submit_to_s3_only` parameter is now deprecated. Insights cost data will now always be uploaded to Dagster Insights via S3."
        )

    partition_end_offset_hrs = -abs(partition_end_offset_hrs)

    partitions_def = HourlyPartitionsDefinition(
        start_date=start_date, end_offset=partition_end_offset_hrs
    )

    @asset(
        name=name,
        group_name=group_name,
        partitions_def=partitions_def,
        required_resource_keys={snowflake_resource_key},
        io_manager_def=fs_io_manager,
    )
    def poll_snowflake_query_history_hour(
        context: AssetExecutionContext,
    ) -> None:
        snowflake: "SnowflakeConnection" = getattr(context.resources, snowflake_resource_key)

        start_hour = context.partition_time_window.start
        end_hour = context.partition_time_window.end

        now = datetime.now().astimezone(timezone.utc)
        earliest_call_time = end_hour + timedelta(minutes=snowflake_usage_latency)
        if now < earliest_call_time:
            err = (
                "Attempted to gather Snowflake usage information before the Snowflake query_history table may be"
                f" available. For hour starting {start_hour.isoformat()} you can call it"
                f" starting at {earliest_call_time.isoformat()} (it is currently"
                f" {now.isoformat()})"
            )
            if allow_partial_partitions:
                context.log.error(err)
            else:
                raise RuntimeError(err)

        costs = (
            get_cost_data_for_hour(
                snowflake,
                start_hour,
                end_hour,
            )
            or []
        )
        snowflake_query_end_time = datetime.now().astimezone(timezone.utc)
        context.log.info(
            f"Fetched query history information from {start_hour.isoformat()} to {end_hour.isoformat()} in {(snowflake_query_end_time - now).total_seconds()} seconds"
        )

        if dry_run:
            pprint(costs)
        else:
            context.log.info(
                f"Submitting cost information for {len(costs)} queries to Dagster Insights"
            )
            put_cost_information(
                context=context,
                metric_name="snowflake_credits",
                cost_information=costs,
                start=start_hour.timestamp(),
                end=end_hour.timestamp(),
            )

    insights_job = define_asset_job(
        job_name,
        AssetSelection.assets(poll_snowflake_query_history_hour),
        partitions_def=partitions_def,
    )

    if schedule_batch_size_hrs == 1:
        insights_schedule = build_schedule_from_partitioned_job(
            job=insights_job,
            minute_of_hour=59,
        )
    else:

        @schedule(
            job=insights_job,
            name=f"{job_name}_schedule_{schedule_batch_size_hrs}_hrs",
            cron_schedule=f"59 0/{schedule_batch_size_hrs} * * *",
        )
        def _insights_schedule(context: ScheduleEvaluationContext):
            timestamp = context.scheduled_execution_time.replace(
                minute=0, second=0, microsecond=0
            ) + timedelta(hours=partition_end_offset_hrs)
            n_hours_ago = timestamp - timedelta(hours=schedule_batch_size_hrs)
            window = TimeWindow(start=n_hours_ago, end=timestamp)

            partition_keys = partitions_def.get_partition_keys_in_time_window(window)

            for partition_key in partition_keys:
                yield insights_job.run_request_for_partition(
                    partition_key=partition_key,
                    run_key=partition_key,
                )

        insights_schedule = _insights_schedule

    # schedule may be a UnresolvedPartitionedAssetScheduleDefinition so we ignore the type check
    return SnowflakeInsightsDefinitions(
        assets=[poll_snowflake_query_history_hour],
        schedule=insights_schedule,  # type: ignore
    )



GET_USER_CLOUD_REQUESTS_QUERY = """
    mutation GetUserCloudRequests($forBranchDeployments: Boolean $forFullDeployments: [String!], $limit: Int, $agentQueues: [String]) {
        userCloudAgent {
            popUserCloudAgentRequests(limit: $limit, forBranchDeployments: $forBranchDeployments, forFullDeployments: $forFullDeployments, agentQueues: $agentQueues) {
                requestId
                requestApi
                requestBody
                deploymentName
                isBranchDeployment
            }
        }
    }

"""

DEPLOYMENTS_QUERY = """
    query Deployments($deploymentNames: [String!]!) {
        deployments(deploymentNames: $deploymentNames) {
            deploymentName
        }
    }
"""

WORKSPACE_ENTRIES_QUERY = """
    query WorkspaceEntries($deploymentNames: [String!]!, $includeAllServerlessDeployments: Boolean!, $agentQueues: [String]) {
        deployments(deploymentNames: $deploymentNames, includeAllServerlessDeployments: $includeAllServerlessDeployments) {
            deploymentName
            isBranchDeployment
            workspaceEntries(agentQueues: $agentQueues) {
                locationName
                serializedDeploymentMetadata
                hasOutdatedData
                metadataTimestamp
            }
        }
    }
"""


ADD_AGENT_HEARTBEATS_MUTATION = """
    mutation AddAgentHeartbeats($serializedAgentHeartbeats: [AgentHeartbeatInput!], $serializedErrors: [String!]) {
        userCloudAgent {
            addAgentHeartbeats (serializedAgentHeartbeats: $serializedAgentHeartbeats, serializedErrors: $serializedErrors) {
                ok
            }
        }
    }
"""

GET_AGENTS_QUERY = """
    query Agents($heartbeatedSince: Float) {
        agents(heartbeatedSince: $heartbeatedSince) {
            id
            status
        }
    }
"""



from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING, Iterable, Iterator, Optional, Union

import yaml
from dagster import (
    AssetCheckResult,
    AssetExecutionContext,
    AssetKey,
    AssetMaterialization,
    AssetObservation,
    OpExecutionContext,
    Output,
)
from dagster_dbt import DbtCliInvocation
from dagster_dbt.version import __version__ as dagster_dbt_version
from packaging import version

from ..insights_utils import extract_asset_info_from_event, handle_raise_on_error
from .bigquery_utils import build_bigquery_cost_metadata, marker_asset_key_for_job

if TYPE_CHECKING:
    from dbt.adapters.base.impl import BaseAdapter
    from google.cloud import bigquery

OPAQUE_ID_SQL_SIGIL = "bigquery_dagster_dbt_v1_opaque_id"
DEFAULT_BQ_REGION = "region-us"
MIN_DAGSTER_DBT_VERSION = "1.7.0"

BIGQUERY_COST_ERROR_MESSAGE = "Could not query information_schema for BigQuery cost information"


@dataclass
class BigQueryCostInfo:
    asset_key: AssetKey
    partition: Optional[str]
    job_id: Optional[str]
    slots_ms: int
    bytes_billed: int

    @property
    def asset_partition_key(self) -> str:
        return (
            f"{self.asset_key.to_string()}:{self.partition}"
            if self.partition
            else self.asset_key.to_string()
        )


@handle_raise_on_error("dbt_cli_invocation")
def dbt_with_bigquery_insights(
    context: Union[OpExecutionContext, AssetExecutionContext],
    dbt_cli_invocation: DbtCliInvocation,
    dagster_events: Optional[
        Iterable[Union[Output, AssetMaterialization, AssetObservation, AssetCheckResult]]
    ] = None,
    skip_config_check=False,
    record_observation_usage: bool = True,
) -> Iterator[Union[Output, AssetMaterialization, AssetObservation, AssetCheckResult]]:
    """Wraps a dagster-dbt invocation to associate each BigQuery query with the produced
    asset materializations. This allows the cost of each query to be associated with the asset
    materialization that it produced.

    If called in the context of an op (rather than an asset), filters out any Output events
    which do not correspond with any output of the op.

    Args:
        context (AssetExecutionContext): The context of the asset that is being materialized.
        dbt_cli_invocation (DbtCliInvocation): The invocation of the dbt CLI to wrap.
        dagster_events (Optional[Iterable[Union[Output, AssetObservation, AssetCheckResult]]]):
            The events that were produced by the dbt CLI invocation. If not provided, it is assumed
            that the dbt CLI invocation has not yet been run, and it will be run and the events
            will be streamed.
        record_observation_usage (bool): If True, associates the usage associated with
            asset observations with that asset. Default is True.

    **Example:**

    .. code-block:: python

        @dbt_assets(manifest=DBT_MANIFEST_PATH)
        def jaffle_shop_dbt_assets(
            context: AssetExecutionContext,
            dbt: DbtCliResource,
        ):
            dbt_cli_invocation = dbt.cli(["build"], context=context)
            yield from dbt_with_bigquery_insights(context, dbt_cli_invocation)
    """
    if not skip_config_check:
        adapter_type = dbt_cli_invocation.manifest["metadata"]["adapter_type"]
        if adapter_type != "bigquery":
            raise RuntimeError(
                f"The 'bigquery' adapter must be used but instead found '{adapter_type}'"
            )
        dbt_project_config = yaml.safe_load(
            (dbt_cli_invocation.project_dir / "dbt_project.yml").open("r")
        )
        # sanity check that the sigil is present somewhere in the query comment
        query_comment = dbt_project_config.get("query-comment")
        if query_comment is None:
            raise RuntimeError("query-comment is required in dbt_project.yml but it was missing")
        comment = query_comment.get("comment")
        if comment is None:
            raise RuntimeError(
                "query-comment.comment is required in dbt_project.yml but it was missing"
            )
        if OPAQUE_ID_SQL_SIGIL not in comment:
            raise RuntimeError(
                "query-comment.comment in dbt_project.yml must contain the string"
                f" '{OPAQUE_ID_SQL_SIGIL}'. Read the Dagster Insights docs for more info."
            )

    if dagster_events is None:
        dagster_events = dbt_cli_invocation.stream()

    asset_info_by_unique_id = {}
    for dagster_event in dagster_events:
        if isinstance(
            dagster_event, (AssetMaterialization, AssetObservation, Output, AssetCheckResult)
        ):
            unique_id = dagster_event.metadata["unique_id"].value
            asset_key, partition = extract_asset_info_from_event(
                context, dagster_event, record_observation_usage
            )
            if not asset_key:
                asset_key = marker_asset_key_for_job(context.job_def)
            asset_info_by_unique_id[unique_id] = (asset_key, partition)

        yield dagster_event

    marker_asset_key = marker_asset_key_for_job(context.job_def)
    run_results_json = dbt_cli_invocation.get_artifact("run_results.json")
    invocation_id = run_results_json["metadata"]["invocation_id"]

    # backcompat-proof in case the invocation does not have an instantiated adapter on it
    adapter: Optional["BaseAdapter"] = getattr(dbt_cli_invocation, "adapter", None)
    if not adapter:
        if version.parse(dagster_dbt_version) < version.parse(MIN_DAGSTER_DBT_VERSION):
            upgrade_message = f" Extracting cost information requires dagster_dbt>={MIN_DAGSTER_DBT_VERSION} (found {dagster_dbt_version}). "
        else:
            upgrade_message = ""

        context.log.error(
            "Could not find a BigQuery adapter on the dbt CLI invocation. Skipping cost analysis."
            + upgrade_message
        )
        return

    cost_by_asset = defaultdict(list)
    try:
        with adapter.connection_named("dagster_insights:bigquery_cost"):
            client: "bigquery.Client" = adapter.connections.get_thread_connection().handle  # pyright: ignore[reportAssignmentType]
            if client.location and client.project:
                # we should populate the location/project from the client, and use that to determine
                # the correct INFORMATION_SCHEMA.JOBS table to query for cost information
                location = client.location
                project = client.project
            else:
                # try fetching the default dataset from the schema, if it exists
                dataset = client.get_dataset(adapter.config.credentials.schema)
                location = dataset.location if dataset else None
                project = client.project or dataset.project

            if not location:
                context.log.exception(f"{BIGQUERY_COST_ERROR_MESSAGE}: Missing location.")
                return

            if not project:
                context.log.exception(f"{BIGQUERY_COST_ERROR_MESSAGE}: Missing project.")

            query_result = client.query(
                rf"""
                    SELECT
                    job_id,
                    regexp_extract(query, r"{OPAQUE_ID_SQL_SIGIL}\[\[\[(.*?):{invocation_id}\]\]\]") as unique_id,
                    total_bytes_billed AS bytes_billed,
                    total_slot_ms AS slots_ms
                    FROM `{project}`.`region-{location.lower()}`.INFORMATION_SCHEMA.JOBS
                    WHERE query like '%{invocation_id}%'
                """
            )
            for row in query_result:
                if not row.unique_id:
                    continue
                asset_key, partition = asset_info_by_unique_id.get(
                    row.unique_id, (marker_asset_key, None)
                )
                if row.bytes_billed or row.slots_ms:
                    cost_info = BigQueryCostInfo(
                        asset_key, partition, row.job_id, row.slots_ms, row.bytes_billed
                    )
                    cost_by_asset[cost_info.asset_partition_key].append(cost_info)
    except:
        context.log.exception(BIGQUERY_COST_ERROR_MESSAGE)
        return

    for cost_info_list in cost_by_asset.values():
        bytes_billed = sum(item.bytes_billed or 0 for item in cost_info_list)
        slots_ms = sum(item.slots_ms or 0 for item in cost_info_list)
        job_ids = [item.job_id for item in cost_info_list if item.job_id]
        asset_key = cost_info_list[0].asset_key
        partition = cost_info_list[0].partition
        yield AssetObservation(
            asset_key=asset_key,
            partition=partition,
            metadata=build_bigquery_cost_metadata(job_ids, bytes_billed, slots_ms),
        )



from typing import Dict, List, Optional

from opentelemetry.metrics import MeterProvider as APIMeterProvider
from opentelemetry.sdk.metrics.export import MetricExporter, MetricReader

from dagster_cloud.opentelemetry.config.exporter import metrics_instrument_types_enum
from dagster_cloud.opentelemetry.enum import MetricsExporterEnum, MetricsReaderEnum


def _validate_dict_keys_as_metric_instruments(src: dict, dict_name: str) -> None:
    """Validate that all keys in a dictionary are valid metric instrument types."""
    for key in src:
        if key not in metrics_instrument_types_enum.enum_values:
            raise ValueError(
                f"OpenTelemetry configuration error - Invalid instrument type '{key}' in {dict_name}."
            )


def build_metric_exporter(exporter_config: dict) -> MetricExporter:
    """Build an OpenTelemetry metric exporter.
    params:
        exporter: dict: The exporter configuration.
          - type: str: The exporter type.
          - params: dict: The exporter parameters (varies).

    Returns:
        MetricExporter: The metric exporter instance.
    """
    exporter_type = exporter_config.get("type", MetricsExporterEnum.ConsoleMetricExporter.value)
    exporter_params = exporter_config.get("params", {})

    if "preferred_temporality" in exporter_params:
        _validate_dict_keys_as_metric_instruments(
            exporter_params["preferred_temporality"], "preferred_temporality"
        )

    if "preferred_aggregation" in exporter_params:
        _validate_dict_keys_as_metric_instruments(
            exporter_params["preferred_aggregation"], "preferred_aggregation"
        )

    if exporter_type == MetricsExporterEnum.HttpProtoOTLPExporter.value:
        from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter

        return OTLPMetricExporter(**exporter_params)
    elif exporter_type == MetricsExporterEnum.GrpcProtoOTLPExporter.value:
        from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter

        return OTLPMetricExporter(**exporter_params)
    elif exporter_type == MetricsExporterEnum.ConsoleMetricExporter.value:
        from opentelemetry.sdk.metrics.export import ConsoleMetricExporter

        return ConsoleMetricExporter(**exporter_params)

    raise ValueError(f"Invalid exporter type: {exporter_type}")


def build_metric_reader(
    exporter: Optional[MetricExporter] = None, reader_config: Optional[dict] = None
) -> MetricReader:
    """Build an OpenTelemetry metric reader.
    params:
        reader: dict: The reader configuration.

    Returns:
        MetricReader: The metric reader instance.
    """
    reader_config = reader_config or {}
    reader_type = reader_config.get("type", MetricsReaderEnum.PeriodicExportingMetricReader.value)
    reader_params = reader_config.get("params", {})

    if reader_type == MetricsReaderEnum.PeriodicExportingMetricReader.value:
        from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader

        if exporter is None:
            raise ValueError("Exporter is required for PeriodicExportingMetricReader.")
        return PeriodicExportingMetricReader(exporter=exporter, **reader_params)
    elif reader_type == MetricsReaderEnum.InMemoryMetricReader.value:
        from opentelemetry.sdk.metrics.export import InMemoryMetricReader

        return InMemoryMetricReader(**reader_params)

    raise ValueError(f"Invalid metric reader type: {reader_type}")


def build_meter_provider(
    metric_readers: List[MetricReader],
    resource_attributes: Optional[Dict[str, str]] = None,
) -> APIMeterProvider:
    """Build an OpenTelemetry meter provider.
    params:
        metric_reader: MetricReader: The metric reader instance.
        resource: Resource (optional): The resource instance.

    Returns:
        MeterProvider: The meter provider instance.
    """
    from opentelemetry.sdk.metrics import MeterProvider
    from opentelemetry.sdk.resources import Resource

    resource = (
        Resource.create(attributes=resource_attributes)
        if resource_attributes
        else Resource.create({})
    )

    return MeterProvider(
        metric_readers=metric_readers,
        resource=resource,
        shutdown_on_exit=False,
        # TODO - implement support for views?
        # https://opentelemetry-python.readthedocs.io/en/latest/_modules/opentelemetry/sdk/metrics/_internal/view.html#View
        # views=views,
    )


def build_noop_meter_provider() -> APIMeterProvider:
    """Build a no-op OpenTelemetry meter provider."""
    from opentelemetry.metrics import NoOpMeterProvider

    return NoOpMeterProvider()



from dagster._core.storage.tags import (
    AUTO_MATERIALIZE_TAG,
    AUTO_OBSERVE_TAG,
    BACKFILL_ID_TAG,
    HIDDEN_TAG_PREFIX,
    PARTITION_NAME_TAG,
    PARTITION_SET_TAG,
    REPOSITORY_LABEL_TAG,
    ROOT_RUN_ID_TAG,
    SCHEDULE_NAME_TAG,
    SENSOR_NAME_TAG,
    SYSTEM_TAG_PREFIX,
)

CLOUD_SYSTEM_TAG_PREFIX = "dagster-cloud/"

PEX_METADATA_TAG = f"{HIDDEN_TAG_PREFIX}pex_metadata"
PEX_TAG_TAG = f"{SYSTEM_TAG_PREFIX}pex_tag"
IGNORE_ALERTS_TAG = f"{CLOUD_SYSTEM_TAG_PREFIX}ignore-alerts"

RUN_TAG_RUN_COLUMN_TAG_KEYS = {
    AUTO_MATERIALIZE_TAG,
    AUTO_OBSERVE_TAG,
    REPOSITORY_LABEL_TAG,
    ROOT_RUN_ID_TAG,
    PARTITION_NAME_TAG,
    PARTITION_SET_TAG,
    SCHEDULE_NAME_TAG,
    SENSOR_NAME_TAG,
    BACKFILL_ID_TAG,
}

DID_ALERT_TAG = f"{CLOUD_SYSTEM_TAG_PREFIX}triggered_alert"
TRIGGERED_ALERT_POLICY_TAG_PREFIX = f"{HIDDEN_TAG_PREFIX}triggered_alert_policy/"
TRIGGERED_ALERT_ID_TAG_PREFIX = f"{HIDDEN_TAG_PREFIX}triggered_alert/"


def get_triggered_alert_policy_key(alert_policy_id: str) -> str:
    return f"{TRIGGERED_ALERT_POLICY_TAG_PREFIX}{alert_policy_id}"


def get_triggered_alert_id_key(alert_id: str) -> str:
    return f"{TRIGGERED_ALERT_ID_TAG_PREFIX}{alert_id}"


def get_policy_names_from_tag_value(policies_str):
    """From a comma-delineated string (whitespace allowed), retrieve alert policy names."""
    return [policy_name.strip() for policy_name in policies_str.split(",")]


def should_tag_be_used_for_indexing_filtering(tag_key: str) -> bool:
    """Determines whether a tag should be added to the run_tags table,
    which is used for indexing and filtering runs.
    Tags not added to the table will still be stored on the run object, but
    will not be directly queryable via the run_tags table.
    """
    if tag_key in RUN_TAG_RUN_COLUMN_TAG_KEYS:
        return False

    if tag_key.startswith(TRIGGERED_ALERT_ID_TAG_PREFIX) or tag_key.startswith(
        TRIGGERED_ALERT_POLICY_TAG_PREFIX
    ):
        return False

    return True



from typing import Any, Dict, Optional

from dagster._core.secrets import SecretsLoader
from dagster._serdes import ConfigurableClass
from dagster._serdes.config_class import ConfigurableClassData
from typing_extensions import Self

SECRETS_QUERY = """
query Secrets($locationName: String) {
    scopedSecrets(locationName: $locationName) {
        secretName
        secretValue
        locationNames
    }
}
"""


class DagsterCloudSecretsLoader(SecretsLoader, ConfigurableClass):
    def __init__(
        self,
        inst_data=None,
    ):
        self._inst_data = inst_data

    def _execute_query(self, query, variables=None):
        return self._instance.graphql_client.execute(query, variable_values=variables)  # pyright: ignore[reportAttributeAccessIssue]

    def get_secrets_for_environment(self, location_name: Optional[str]) -> Dict[str, str]:
        res = self._execute_query(
            SECRETS_QUERY,
            variables={"locationName": location_name},
        )

        secrets = res["data"]["scopedSecrets"]

        # Place secrets scoped to this location at the end so that they take priority over secrets
        # with the same name but no location scopes
        secrets = sorted(secrets, key=lambda secret: "1" if secret["locationNames"] else "0")

        return {secret["secretName"]: secret["secretValue"] for secret in secrets}

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @classmethod
    def from_config_value(cls, inst_data: ConfigurableClassData, config_value: Any) -> Self:
        return cls(inst_data=inst_data, **config_value)



from contextlib import contextmanager, nullcontext
from typing import Iterator, List

from dagster import AssetObservation
from dagster._annotations import experimental
from dagster_gcp import BigQueryResource
from dagster_gcp.bigquery.utils import setup_gcp_creds
from google.cloud import bigquery

from dagster_cloud.dagster_insights.insights_utils import get_current_context_and_asset_key

from .bigquery_utils import build_bigquery_cost_metadata, marker_asset_key_for_job

OUTPUT_NON_ASSET_SIGIL = "__bigquery_query_metadata_"


class WrappedBigQueryClient(bigquery.Client):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._query_bytes = []
        self._query_slots_ms = []
        self._job_ids = []

    def query(self, *args, **kwargs) -> bigquery.QueryJob:
        bq_job = super().query(*args, **kwargs)
        self._query_bytes.append(bq_job.total_bytes_billed or 0)
        self._query_slots_ms.append(bq_job.slot_millis or 0)
        self._job_ids.append(bq_job.job_id)
        return bq_job

    @property
    def job_ids(self) -> List[str]:
        return self._job_ids

    @property
    def total_bytes_billed(self) -> int:
        return sum([x for x in self._query_bytes])

    @property
    def total_slots_ms(self) -> int:
        return sum([x for x in self._query_slots_ms])


@experimental
class InsightsBigQueryResource(BigQueryResource):
    """A wrapper around :py:class:`BigQueryResource` which automatically collects metadata about
    BigQuery costs which can be attributed to Dagster jobs and assets.

    A simple example of loading data into BigQuery and subsequently querying that data is shown below:

    Examples:
        .. code-block:: python

            from dagster import job, op
            from dagster_gcp import BigQueryResource
            from dagster_insights import InsightsBigQueryResource

            @op
            def get_one(bigquery_resource: BigQueryResource):
                with bigquery_resource.get_client() as client:
                    client.query("SELECT * FROM my_dataset.my_table")

            @job
            def my_bigquery_job():
                get_one()

            my_bigquery_job.execute_in_process(
                resources={ "bigquery": InsightsBigQueryResource(project="my-project") }
            )
    """

    @contextmanager
    def get_client(self) -> Iterator[bigquery.Client]:
        context, asset_key = get_current_context_and_asset_key()

        with setup_gcp_creds(self.gcp_credentials) if self.gcp_credentials else nullcontext():
            client = WrappedBigQueryClient(project=self.project, location=self.location)

            yield client

            if not asset_key:
                asset_key = marker_asset_key_for_job(context.job_def)

            if client.total_bytes_billed or client.total_slots_ms:
                context.log_event(
                    AssetObservation(
                        asset_key=asset_key,
                        metadata=build_bigquery_cost_metadata(
                            client.job_ids, client.total_bytes_billed, client.total_slots_ms
                        ),
                    )
                )
